import attr
import traceback
from collections import namedtuple

from .. import repositories, entities, services


class ExecutionStatus:
    SUCCESS = "success"
    FAILED = "failed"
    IN_PROGRESS = "inProgress"
    CREATED = "created"


@attr.s
class Execution(entities.BaseEntity):
    """
    Service execution entity
    """
    # platform
    id = attr.ib()
    url = attr.ib(repr=False)
    creator = attr.ib()
    createdAt = attr.ib()
    updatedAt = attr.ib(repr=False)
    input = attr.ib()
    output = attr.ib(repr=False)
    feedbackQueue = attr.ib(repr=False)
    status = attr.ib(repr=False)
    syncReplyTo = attr.ib(repr=False)
    latest_status = attr.ib()
    function_name = attr.ib()
    duration = attr.ib()
    attempts = attr.ib()
    max_attempts = attr.ib()
    to_terminate = attr.ib(type=bool)

    # name changed
    trigger_id = attr.ib()
    service_id = attr.ib()
    pipeline_id = attr.ib()
    project_id = attr.ib()

    # sdk
    _client_api = attr.ib(type=services.ApiClient, repr=False)
    _service = attr.ib(repr=False)
    _project = attr.ib(repr=False, default=None)
    _repositories = attr.ib(repr=False)

    ################
    # repositories #
    ################
    @_repositories.default
    def set_repositories(self):
        reps = namedtuple('repositories',
                          field_names=['executions', 'services'])

        if self._project is not None:
            services_repo = self._project.services
            executions_repo = self._project.executions
        elif self._service is not None:
            services_repo = self._service.services
            executions_repo = self._service.executions
        else:
            services_repo = repositories.Services(client_api=self._client_api,
                                                  project=self._project,
                                                  package=None)
            executions_repo = repositories.Executions(client_api=self._client_api,
                                                      project=self._project,
                                                      service=self._service)

        r = reps(executions=executions_repo,
                 services=services_repo)
        return r

    @property
    def services(self):
        assert isinstance(self._repositories.services, repositories.Services)
        return self._repositories.services

    @property
    def executions(self):
        assert isinstance(self._repositories.executions, repositories.Executions)
        return self._repositories.executions

    @staticmethod
    def _protected_from_json(_json, client_api, service=None, is_fetched=True):
        """
        Same as from_json but with try-except to catch if error
        :param _json:
        :param client_api:
        :return:
        """
        try:
            execution = Execution.from_json(_json=_json,
                                            client_api=client_api,
                                            service=service,
                                            is_fetched=is_fetched)
            status = True
        except Exception:
            execution = traceback.format_exc()
            status = False
        return status, execution

    @classmethod
    def from_json(cls, _json, client_api, service=None, is_fetched=True):
        inst = cls(
            feedbackQueue=_json.get('feedbackQueue', None),
            service_id=_json.get('serviceId', None),
            project_id=_json.get('projectId', None),
            pipeline_id=_json.get('pipelineId', None),
            latest_status=_json.get('latestStatus', None),
            syncReplyTo=_json.get('syncReplyTo', None),
            createdAt=_json.get('createdAt', None),
            updatedAt=_json.get('updatedAt', None),
            creator=_json.get('creator', None),
            trigger_id=_json.get('triggerId', None),
            attempts=_json.get('attempts', None),
            max_attempts=_json.get('maxAttempts', None),
            output=_json.get('output', None),
            status=_json.get('status', None),
            duration=_json.get('duration', None),
            function_name=_json.get('functionName', entities.package_defaults.DEFAULT_PACKAGE_FUNCTION_NAME),
            input=_json.get('input', None),
            url=_json.get('url', None),
            id=_json.get('id', None),
            to_terminate=_json.get('toTerminate', False),
            client_api=client_api,
            service=service
        )
        inst.is_fetched = is_fetched
        return inst

    def to_json(self):
        """
        Returns platform _json format of object

        :return: platform json format of object
        """
        # get excluded
        _json = attr.asdict(self, filter=attr.filters.exclude(attr.fields(Execution)._client_api,
                                                              attr.fields(Execution)._service,
                                                              attr.fields(Execution)._project,
                                                              attr.fields(Execution).to_terminate,
                                                              attr.fields(Execution)._repositories,
                                                              attr.fields(Execution).project_id,
                                                              attr.fields(Execution).service_id,
                                                              attr.fields(Execution).pipeline_id,
                                                              attr.fields(Execution).trigger_id,
                                                              attr.fields(Execution).function_name,
                                                              attr.fields(Execution).max_attempts,
                                                              attr.fields(Execution).latest_status))
        # rename
        _json['projectId'] = self.project_id
        _json['triggerId'] = self.trigger_id
        _json['serviceId'] = self.service_id
        _json['pipelineId'] = self.pipeline_id
        _json['functionName'] = self.function_name
        _json['latestStatus'] = self.latest_status
        _json['maxAttempts'] = self.max_attempts
        _json['toTerminate'] = self.to_terminate
        return _json

    @property
    def service(self):
        if self._service is None:
            self._service = self.services.get(service_id=self.service_id, fetch=None)
        assert isinstance(self._service, entities.Service)
        return self._service

    @property
    def project(self):
        if self._project is None:
            self._project = repositories.Projects(client_api=self._client_api).get(project_id=self.project_id,
                                                                                   fetch=None)
        assert isinstance(self._project, entities.Project)
        return self._project

    def get_latest_status(self):
        self.latest_status = self.executions.get(execution_id=self.id).latest_status
        return self.latest_status

    @service.setter
    def service(self, service):
        if not isinstance(service, entities.Service):
            raise ValueError('Must input a valid service entity')
        self._service = service

    def progress_update(self, status=None, percent_complete=None, message=None, output=None):
        """
        Update Execution Progress

        :param status:
        :param percent_complete:
        :param message:
        :param output:
        :return:
        """
        return self.executions.progress_update(execution_id=self.id,
                                               status=status,
                                               percent_complete=percent_complete,
                                               message=message,
                                               output=output)

    def update(self):
        """
        Update execution changes to platform
        :return: execution entity
        """
        return self.executions.update(execution=self)

    def logs(self):
        """
        Print logs for execution
        """
        self.services.log(execution_id=self.id, view=True, service=self.service)

    def increment(self):
        """
        Increment attempts

        :return:
        """
        self.attempts = self.executions.increment(execution=self)

    def terminate(self):
        """
        Terminate execution

        :return:
        """
        return self.executions.terminate(execution=self)

    def wait(self):
        """
        Wait for execution

        :return:
        """
        return self.executions.wait(execution_id=self.id)



from collections import namedtuple
import traceback
import logging
import attr

from .. import repositories, miscellaneous, services, entities

logger = logging.getLogger(name=__name__)


@attr.s()
class Project(entities.BaseEntity):
    """
    Project entity
    """

    _contributors = attr.ib(repr=False)
    createdAt = attr.ib()
    creator = attr.ib()
    id = attr.ib()
    name = attr.ib()
    org = attr.ib(repr=False)
    updatedAt = attr.ib(repr=False)
    role = attr.ib(repr=False)
    account = attr.ib(repr=False)

    # name change
    feature_constraints = attr.ib()

    # api
    _client_api = attr.ib(type=services.ApiClient, repr=False)

    # repositories
    _repositories = attr.ib(repr=False)

    @_repositories.default
    def set_repositories(self):
        reps = namedtuple('repositories',
                          'projects triggers datasets items recipes packages codebases artifacts times_series services '
                          'executions assignments tasks bots webhooks models analytics')
        datasets = repositories.Datasets(client_api=self._client_api, project=self)

        r = reps(projects=repositories.Projects(client_api=self._client_api),
                 webhooks=repositories.Webhooks(client_api=self._client_api, project=self),
                 items=repositories.Items(client_api=self._client_api, datasets=datasets),
                 recipes=repositories.Recipes(client_api=self._client_api, project=self, project_id=self.id),
                 datasets=datasets,
                 executions=repositories.Executions(client_api=self._client_api, project=self),
                 triggers=repositories.Triggers(client_api=self._client_api, project=self),
                 packages=repositories.Packages(project=self, client_api=self._client_api),
                 models=repositories.Models(project=self, client_api=self._client_api),
                 codebases=repositories.Codebases(project=self, client_api=self._client_api),
                 artifacts=repositories.Artifacts(project=self, client_api=self._client_api),
                 times_series=repositories.TimesSeries(project=self, client_api=self._client_api),
                 services=repositories.Services(client_api=self._client_api, project=self),
                 assignments=repositories.Assignments(project=self, client_api=self._client_api),
                 tasks=repositories.Tasks(client_api=self._client_api, project=self),
                 bots=repositories.Bots(client_api=self._client_api, project=self),
                 analytics=repositories.Analytics(client_api=self._client_api, project=self)
                 )
        return r

    @property
    def triggers(self):
        assert isinstance(self._repositories.triggers, repositories.Triggers)
        return self._repositories.triggers

    @property
    def items(self):
        assert isinstance(self._repositories.items, repositories.Items)
        return self._repositories.items

    @property
    def recipes(self):
        assert isinstance(self._repositories.recipes, repositories.Recipes)
        return self._repositories.recipes

    @property
    def services(self):
        assert isinstance(self._repositories.services, repositories.Services)
        return self._repositories.services

    @property
    def executions(self):
        assert isinstance(self._repositories.executions, repositories.Executions)
        return self._repositories.executions

    @property
    def projects(self):
        assert isinstance(self._repositories.projects, repositories.Projects)
        return self._repositories.projects

    @property
    def datasets(self):
        assert isinstance(self._repositories.datasets, repositories.Datasets)
        return self._repositories.datasets

    @property
    def packages(self):
        assert isinstance(self._repositories.packages, repositories.Packages)
        return self._repositories.packages

    @property
    def models(self):
        assert isinstance(self._repositories.models, repositories.Models)
        return self._repositories.models

    @property
    def codebases(self):
        assert isinstance(self._repositories.codebases, repositories.Codebases)
        return self._repositories.codebases

    @property
    def webhooks(self):
        assert isinstance(self._repositories.webhooks, repositories.Webhooks)
        return self._repositories.webhooks

    @property
    def artifacts(self):
        assert isinstance(self._repositories.artifacts, repositories.Artifacts)
        return self._repositories.artifacts

    @property
    def times_series(self):
        assert isinstance(self._repositories.times_series, repositories.TimesSeries)
        return self._repositories.times_series

    @property
    def assignments(self):
        assert isinstance(self._repositories.assignments, repositories.Assignments)
        return self._repositories.assignments

    @property
    def tasks(self):
        assert isinstance(self._repositories.tasks, repositories.Tasks)
        return self._repositories.tasks

    @property
    def bots(self):
        assert isinstance(self._repositories.bots, repositories.Bots)
        return self._repositories.bots

    @property
    def analytics(self):
        assert isinstance(self._repositories.analytics, repositories.Analytics)
        return self._repositories.analytics

    @property
    def contributors(self):
        return miscellaneous.List([entities.User.from_json(_json=_json,
                                                           client_api=self._client_api,
                                                           project=self) for _json in self._contributors])

    @staticmethod
    def _protected_from_json(_json, client_api):
        """
        Same as from_json but with try-except to catch if error
        :param _json:
        :param client_api:
        :return:
        """
        try:
            project = Project.from_json(_json=_json,
                                        client_api=client_api)
            status = True
        except Exception:
            project = traceback.format_exc()
            status = False
        return status, project

    @classmethod
    def from_json(cls, _json, client_api, is_fetched=True):
        """
        Build a Project entity object from a json

        :param is_fetched: is Entity fetched from Platform
        :param _json: _json response from host
        :param client_api: client_api
        :return: Project object
        """
        inst = cls(feature_constraints=_json.get('featureConstraints', None),
                   contributors=_json.get('contributors', None),
                   createdAt=_json.get('createdAt', None),
                   updatedAt=_json.get('updatedAt', None),
                   creator=_json.get('creator', None),
                   account=_json.get('account', None),
                   name=_json.get('name', None),
                   role=_json.get('role', None),
                   org=_json.get('org', None),
                   id=_json.get('id', None),
                   client_api=client_api)
        inst.is_fetched = is_fetched
        return inst

    def to_json(self):
        """
        Returns platform _json format of object

        :return: platform json format of object
        """
        output_dict = attr.asdict(self,
                                  filter=attr.filters.exclude(attr.fields(Project)._client_api,
                                                              attr.fields(Project)._repositories,
                                                              attr.fields(Project).feature_constraints,
                                                              attr.fields(Project)._contributors))
        output_dict['contributors'] = self._contributors
        output_dict['featureConstraints'] = self.feature_constraints

        return output_dict

    def delete(self, sure=False, really=False):
        """
        Delete the project forever!

        :param sure: are you sure you want to delete?
        :param really: really really?
        :return: True
        """
        return self.projects.delete(project_id=self.id,
                                    sure=sure,
                                    really=really)

    def update(self, system_metadata=False):
        """
        Update the project

        :return: Project object
        """
        return self.projects.update(project=self, system_metadata=system_metadata)

    def checkout(self):
        """
        Checkout the project

        :return:
        """
        self.projects.checkout(project=self)

    def open_in_web(self):
        """
        Open the project in web platform

        :return:
        """
        self.projects.open_in_web(project=self)

    def add_member(self, email, role='engineer'):
        return self.projects.add_member(email=email, role=role, project_id=self.id)



import numpy as np

from . import BaseAnnotationDefinition


class Polygon(BaseAnnotationDefinition):
    """
    Polygon annotation object
    """

    def __init__(self, geo, label, is_open=True, attributes=None):
        self.type = "segment"
        self.geo = geo
        self.label = label
        self.is_open = is_open
        if attributes is None:
            attributes = list()
        self.attributes = attributes

    @property
    def x(self):
        return self.geo[:, 0]

    @property
    def y(self):
        return self.geo[:, 1]

    @property
    def left(self):
        return np.min(self.x)

    @property
    def top(self):
        return np.min(self.y)

    @property
    def right(self):
        return np.max(self.x)

    @property
    def bottom(self):
        return np.max(self.y)

    def to_coordinates(self, color):
        return [[{"x": float(x), "y": float(y)} for x, y in self.geo]]

    @staticmethod
    def from_coordinates(coordinates):
        return np.asarray([[pt["x"], pt["y"]] for pt in coordinates])

    def show(self, image, thickness, with_text, height, width, annotation_format, color):
        """
        Show annotation as ndarray
        :param image: empty or image to draw on
        :param thickness:
        :param with_text: not required
        :param height: item height
        :param width: item width
        :param annotation_format: options: dl.ViewAnnotationOptions.list()
        :param color: color
        :return: ndarray
        """
        try:
            import cv2
        except ImportError:
            self.logger.error(
                'Import Error! Cant import cv2. Annotations operations will be limited. import manually and fix errors')
            raise

        if thickness is None:
            thickness = 2
        elif thickness == -1 and self.is_open:
            thickness = 2
            self.logger.warning('Cannot fill out open polygon, setting default thickness to 2')
        # draw annotation
        if self.is_open:
            image = cv2.polylines(
                img=image,
                pts=[np.round(self.geo).astype(int)],
                color=color,
                isClosed=False,
                thickness=thickness,
            )
        else:
            image = cv2.drawContours(
                image=image,
                contours=[np.round(self.geo).astype(int)],
                contourIdx=-1,
                color=color,
                thickness=thickness,
            )
        if with_text:
            image = self.add_text_to_image(image=image, annotation=self)
        return image

    @classmethod
    def from_segmentation(cls, mask, label, attributes=None, epsilon=None, max_instances=1, min_area=0, is_open=True):
        """
        Convert binary mask to Polygon
        :param mask: binary mask (0,1)
        :param label: annotation label
        :param attributes: annotations list of attributes
        :param epsilon: from opencv: specifying the approximation accuracy. This is the maximum distance
    .   between the original curve and its approximation. if 0 all points are returns
        :param max_instances: number of max instances to return. if None all wil be returned
        :param min_area: remove polygons with area lower thn this threshold (pixels)
        :param is_open: open or close the polygon
        :return: Polygon annotation
        """
        try:
            import cv2
        except ImportError:
            raise ModuleNotFoundError('opencv not found. Must install to perform this function')

        # mask float
        mask = 1. * mask
        # normalize to 1
        mask /= np.max(mask)
        # threshold the mask
        ret, thresh = cv2.threshold(mask, 0.5, 255, 0)
        # find contours
        im2, contours, hierarchy = cv2.findContours(thresh.astype(np.uint8), cv2.RETR_TREE,
                                                    cv2.CHAIN_APPROX_NONE)
        if len(contours) == 0:
            # no contours were found
            new_pts_list = []
        else:
            # calculate contours area
            areas = np.asarray([cv2.contourArea(cnt) for cnt in contours])
            # take onr contour with maximum area
            sorted_areas_inds = areas.argsort()[::-1]
            filtered_contours = [contours[s_ind] for s_ind in
                                 sorted_areas_inds if areas[s_ind] > min_area]  # filter by area size
            filtered_contours = filtered_contours[:max_instances]  # take only the first max_instance of the results
            # estimate contour to reduce number of points
            new_pts_list = list()
            for curve in filtered_contours:
                if epsilon is None:
                    epsilon = 0.0005 * cv2.arcLength(curve=curve,
                                                     closed=not is_open)
                new_pts_list.append(np.squeeze(cv2.approxPolyDP(curve=curve,
                                                                epsilon=epsilon,
                                                                closed=not is_open)))
        polygons = [cls(geo=new_pts,
                        label=label,
                        attributes=attributes,
                        is_open=is_open
                        ) for new_pts in new_pts_list]

        if len(polygons) == 1:
            polygons = polygons[0]
        return polygons

    @classmethod
    def from_json(cls, _json):
        if "coordinates" in _json:
            geo = cls.from_coordinates(coordinates=_json["coordinates"][0])
        elif "data" in _json:
            geo = cls.from_coordinates(coordinates=_json["data"][0])
        else:
            raise ValueError(
                'can not find "coordinates" or "data" in annotation. id: %s'
                % _json["id"]
            )
        is_open = _json.get('is_open', False)
        attributes = _json.get("attributes", list())
        return cls(
            geo=geo,
            label=_json["label"],
            attributes=attributes,
            is_open=is_open
        )



import logging
import json
import attr

from .. import exceptions, entities

logger = logging.getLogger("dataloop.function")


@attr.s
class PackageFunction(entities.BaseEntity):
    """
    Webhook object
    """
    # platform
    display_name = attr.ib(default=None)
    display_icon = attr.ib(repr=False, default=None)
    display_scope = attr.ib(default=None)
    outputs = attr.ib()
    name = attr.ib(default=entities.package_defaults.DEFAULT_PACKAGE_FUNCTION_NAME)
    description = attr.ib(default='')
    inputs = attr.ib()

    @classmethod
    def from_json(cls, _json):
        inputs = [FunctionIO.from_json(_io) for _io in _json.get('input', list())]
        outputs = [FunctionIO.from_json(_io) for _io in _json.get('output', list())]
        return cls(
            description=_json.get("description", None),
            name=_json.get("name", None),
            inputs=inputs,
            outputs=outputs,
            display_icon=_json.get('displayIcon', None),
            display_name=_json.get('displayName', None),
            display_scope=_json.get('displayScope', None),
        )

    @outputs.default
    def get_outputs(self):
        outputs = list()
        return outputs

    @inputs.default
    def get_inputs(self):
        inputs = list()
        return inputs

    def to_json(self):
        _json = attr.asdict(
            self,
            filter=attr.filters.exclude(attr.fields(PackageFunction).inputs,
                                        attr.fields(PackageFunction).outputs,
                                        attr.fields(PackageFunction).display_icon,
                                        attr.fields(PackageFunction).display_name,
                                        attr.fields(PackageFunction).display_scope,
                                        ),
        )
        inputs = self.inputs
        # check in inputs is a list
        if not isinstance(inputs, list):
            inputs = [inputs]
        # if is dtlpy entity convert to dict
        if inputs and isinstance(inputs[0], entities.FunctionIO):
            inputs = [_io.to_json() for _io in inputs]

        outputs = self.outputs
        # check in inputs is a list
        if not isinstance(outputs, list):
            outputs = [outputs]
        # if is dtlpy entity convert to dict
        if outputs and isinstance(outputs[0], entities.FunctionIO):
            outputs = [_io.to_json() for _io in outputs]

        _json['input'] = inputs
        _json['output'] = outputs
        _json['displayName'] = self.display_name
        _json['displayIcon'] = self.display_icon
        _json['displayScope'] = self.display_scope
        return _json


class PackageInputType:
    ITEM = "Item"
    DATASET = "Dataset"
    ANNOTATION = "Annotation"
    JSON = "Json"
    MODEL = "Model"
    SNAPSHOT = "Snapshot"
    PACKAGE = "Package"
    SERVICE = "Service"
    PROJECT = "Project"
    EXECUTION = "Execution"


@attr.s
class FunctionIO:
    INPUT_TYPES = [val for key, val in PackageInputType.__dict__.items() if not key.startswith('_')]
    type = attr.ib(type=str)
    value = attr.ib(default=None)
    name = attr.ib(type=str)

    @name.default
    def set_name(self):
        if self.type == PackageInputType.ITEM:
            return 'item'
        elif self.type == PackageInputType.DATASET:
            return 'dataset'
        elif self.type == PackageInputType.ANNOTATION:
            return 'annotation'
        elif self.type == PackageInputType.PROJECT:
            return 'project'
        elif self.type == PackageInputType.PACKAGE:
            return 'package'
        elif self.type == PackageInputType.SERVICE:
            return 'service'
        elif self.type == PackageInputType.EXECUTION:
            return 'execution'
        elif self.type == PackageInputType.MODEL:
            return 'model'
        elif self.type == PackageInputType.SNAPSHOT:
            return 'snapshot'
        else:
            return 'config'

    # noinspection PyUnusedLocal
    @type.validator
    def check_type(self, attribute, value):
        if value not in self.INPUT_TYPES:
            raise exceptions.PlatformException('400',
                                               'Invalid input type please select from: {}'.format(self.INPUT_TYPES))

    @staticmethod
    def is_json_serializable(val):
        try:
            json.dumps(val)
            is_json_serializable = True
        except Exception:
            is_json_serializable = False
        return is_json_serializable

    # noinspection PyUnusedLocal
    @value.validator
    def check_value(self, attribute, value):
        value_ok = True
        expected_value = 'Expected value should be:'
        if self.type == PackageInputType.JSON:
            expected_value = '{} json serializable'.format(expected_value)
            if not self.is_json_serializable(value):
                value_ok = False
        else:
            expected_value = '{} string or dictionary'.format(expected_value)
            if type(value) not in [dict, str]:
                value_ok = False

        if not value_ok and value is not None:
            raise exceptions.PlatformException('400', 'Illegal value. {}'.format(expected_value))

    def to_json(self, resource='package'):
        if resource == 'package':
            _json = {
                'name': self.name,
                'type': self.type
            }
        elif resource in ['execution', 'service']:
            _json = {
                self.name: self.value
            }
        else:
            raise exceptions.PlatformException('400', 'Please select resource from: package, execution')

        return _json

    @classmethod
    def from_json(cls, _json):
        return cls(
            type=_json.get('type', None),
            value=_json.get('value', None),
            name=_json.get('name', None)
        )



import validators
import traceback
import tempfile
import requests
import asyncio
import logging
import pandas
import shutil
import json
import time
import tqdm
import os
import io
from requests.adapters import HTTPAdapter
from urllib3.util import Retry

from .. import PlatformException, entities, repositories
from ..services import Reporter

logger = logging.getLogger(name=__name__)

NUM_TRIES = 5  # try to upload 3 time before fail on item


class UploadElement:
    def __init__(self,
                 element_type,
                 buffer,
                 remote_filepath,
                 annotations_filepath,
                 remote_name=None,
                 link_dataset_id=None,
                 item_metadata=None):
        self.type = element_type
        self.buffer = buffer
        self.remote_filepath = remote_filepath
        self.remote_name = remote_name
        self.item_metadata = item_metadata
        self.exists_in_remote = False
        self.checked_in_remote = False
        self.annotations_filepath = annotations_filepath
        self.link_dataset_id = link_dataset_id


class Uploader:
    def __init__(self, items_repository: repositories.Items):
        assert isinstance(items_repository, repositories.Items)
        self.items_repository = items_repository
        self.__stop_create_existence_dict = False

    def upload(
            self,
            # what to upload
            local_path,
            local_annotations_path=None,
            # upload options
            remote_path=None,
            remote_name=None,
            file_types=None,
            num_workers=None,
            overwrite=False,
            item_metadata=None
    ):
        """
        Upload local file to dataset.
        Local filesystem will remain.
        If "*" at the end of local_path (e.g. "/images/*") items will be uploaded without head directory

        :param overwrite: optional - default = False
        :param local_path: local file or folder to upload
        :param local_annotations_path: path to dataloop format annotations json files.
        :param remote_path: remote path to save.
        :param remote_name: remote base name to save.
        :param file_types: list of file type to upload. e.g ['.jpg', '.png']. default is all
        :param num_workers: NOT USED deprecated
        :param item_metadata: upload the items with the metadata dictionary
        :return: Output (list)
        """
        ###################
        # Default options #
        ###################
        mode = 'skip'
        if overwrite:
            mode = 'overwrite'
        if num_workers is not None:
            logger.warning('[DeprecationWarning] input argument "num_workers"'
                           ' will be deprecated from upload() after version 1.17.0\n'
                           'To set number of processes use "dtlpy.client_api.num_processes=int()"')
            self.items_repository._client_api.num_processes = num_workers

        if isinstance(local_path, pandas.DataFrame):
            elements = self._build_elements_from_df(local_path)
        else:
            elements = self._build_elements_from_inputs(local_path=local_path,
                                                        local_annotations_path=local_annotations_path,
                                                        # upload options
                                                        remote_path=remote_path,
                                                        remote_name=remote_name,
                                                        file_types=file_types,
                                                        item_metadata=item_metadata)
        num_files = len(elements)
        logger.info("Uploading {} items..".format(num_files))
        reporter = Reporter(num_workers=num_files, resource=Reporter.ITEMS_UPLOAD)
        disable_pbar = self.items_repository._client_api.verbose.disable_progress_bar or num_files == 1
        pbar = tqdm.tqdm(total=num_files, disable=disable_pbar)
        futures = [None for _ in range(num_files)]
        for i_item in range(num_files):
            element = elements[i_item]
            # upload
            futures[i_item] = asyncio.run_coroutine_threadsafe(
                self.__upload_single_item_wrapper(i_item=i_item,
                                                  element=element,
                                                  mode=mode,
                                                  pbar=pbar,
                                                  reporter=reporter),
                loop=self.items_repository._client_api.event_loops('items.upload').loop)

        _ = [future.result() for future in futures]
        pbar.close()
        # summary
        logger.info("Number of total files: {}".format(num_files))
        for action in reporter.status_list:
            n_for_action = reporter.status_count(status=action)
            logger.info("Number of files {}: {}".format(action, n_for_action))

        # log error
        errors_count = reporter.failure_count
        if errors_count > 0:
            log_filepath = reporter.generate_log_files()
            logger.warning("Errors in {n_error} files. See {log_filepath} for full log".format(
                n_error=errors_count, log_filepath=log_filepath))

        return reporter.output

    def _build_elements_from_inputs(self,
                                    local_path,
                                    local_annotations_path,
                                    # upload options
                                    remote_path,
                                    file_types,
                                    remote_name,
                                    item_metadata):
        if remote_path is None:
            remote_path = '/'
        if not remote_path.endswith("/"):
            remote_path += "/"
        if file_types is not None and not isinstance(file_types, list):
            msg = '"file_types" should be a list of file extension. e.g [".jpg", ".png"]'
            raise PlatformException(error="400", message=msg)
        if item_metadata is not None:
            if not isinstance(item_metadata, dict):
                msg = '"item_metadata" should be a metadata dictionary. Got type: {}'.format(type(item_metadata))
                raise PlatformException(error="400", message=msg)

        ##########################
        # Convert inputs to list #
        ##########################
        local_annotations_path_list = None
        remote_name_list = None
        if not isinstance(local_path, list):
            local_path_list = [local_path]
            if remote_name is not None:
                if not isinstance(remote_name, str):
                    raise PlatformException(error="400",
                                            message='remote_name must be a string, got: {}'.format(type(remote_name)))
                remote_name_list = [remote_name]
            if local_annotations_path is not None:
                if not isinstance(local_annotations_path, str):
                    raise PlatformException(error="400",
                                            message='local_annotations_path must be a string, got: {}'.format(
                                                type(local_annotations_path)))
                local_annotations_path_list = [local_annotations_path]
        else:
            local_path_list = local_path
            if remote_name is not None:
                if not isinstance(remote_name, list):
                    raise PlatformException(error="400",
                                            message='remote_name must be a list, got: {}'.format(type(remote_name)))
                if not len(remote_name) == len(local_path_list):
                    raise PlatformException(error="400",
                                            message='remote_name and local_path_list must be of same length. '
                                                    'Received: remote_name: {}, '
                                                    'local_path_list: {}'.format(len(remote_name),
                                                                                 len(local_path_list)))
                remote_name_list = remote_name
            if local_annotations_path is not None:
                if not len(local_annotations_path) == len(local_path_list):
                    raise PlatformException(error="400",
                                            message='local_annotations_path and local_path_list must be of same lenght.'
                                                    ' Received: local_annotations_path: {}, '
                                                    'local_path_list: {}'.format(len(local_annotations_path),
                                                                                 len(local_path_list)))
                local_annotations_path_list = local_annotations_path

        if local_annotations_path is None:
            local_annotations_path_list = [None] * len(local_path_list)

        if remote_name is None:
            remote_name_list = [None] * len(local_path_list)

        elements = list()
        total_size = 0
        for upload_item_element, remote_name, upload_annotations_element in zip(local_path_list, remote_name_list,
                                                                                local_annotations_path_list):
            if isinstance(upload_item_element, str):
                with_head_folder = True
                if upload_item_element.endswith('*'):
                    with_head_folder = False
                    upload_item_element = os.path.dirname(upload_item_element)

                if os.path.isdir(upload_item_element):
                    # create a list of all the items to upload
                    for root, subdirs, files in os.walk(upload_item_element):
                        for filename in files:
                            _, ext = os.path.splitext(filename)
                            if file_types is None or ext in file_types:
                                filepath = os.path.join(root, filename)  # get full image filepath
                                # extract item's size
                                total_size += os.path.getsize(filepath)
                                # get annotations file
                                if upload_annotations_element is not None:
                                    # change path to annotations
                                    annotations_filepath = filepath.replace(upload_item_element,
                                                                            upload_annotations_element)
                                    # remove image extension
                                    annotations_filepath, _ = os.path.splitext(annotations_filepath)
                                    # add json extension
                                    annotations_filepath += ".json"
                                else:
                                    annotations_filepath = None
                                if with_head_folder:
                                    remote_filepath = remote_path + os.path.relpath(filepath, os.path.dirname(
                                        upload_item_element))
                                else:
                                    remote_filepath = remote_path + os.path.relpath(filepath, upload_item_element)
                                element = UploadElement(element_type='file',
                                                        buffer=filepath,
                                                        remote_filepath=remote_filepath.replace("\\", "/"),
                                                        annotations_filepath=annotations_filepath,
                                                        item_metadata=item_metadata)
                                elements.append(element)

                # add single file
                elif os.path.isfile(upload_item_element):
                    filepath = upload_item_element
                    # extract item's size
                    total_size += os.path.getsize(filepath)
                    # determine item's remote base name
                    if remote_name is None:
                        remote_name = os.path.basename(filepath)
                    # get annotations file
                    if upload_annotations_element is not None:
                        # change path to annotations
                        annotations_filepath = filepath.replace(upload_item_element, upload_annotations_element)
                        # remove image extension
                        annotations_filepath, _ = os.path.splitext(annotations_filepath)
                        # add json extension
                        annotations_filepath += ".json"
                    else:
                        annotations_filepath = None
                    # append to list
                    remote_filepath = remote_path + remote_name
                    element = UploadElement(element_type='file',
                                            buffer=filepath,
                                            remote_filepath=remote_filepath,
                                            annotations_filepath=annotations_filepath,
                                            item_metadata=item_metadata)
                    elements.append(element)
                elif self.is_url(upload_item_element):
                    # noinspection PyTypeChecker
                    if remote_name is None:
                        remote_name = str(upload_item_element.split('/')[-1])
                    remote_filepath = remote_path + remote_name

                    element = UploadElement(element_type='url',
                                            buffer=upload_item_element,
                                            remote_filepath=remote_filepath,
                                            annotations_filepath=upload_annotations_element,
                                            item_metadata=item_metadata)
                    elements.append(element)
                else:
                    raise PlatformException("404", "Unknown local path: {}".format(local_path))
            elif isinstance(upload_item_element, entities.Item):
                link = entities.Link(ref=upload_item_element.id, type='id', dataset_id=upload_item_element.datasetId,
                                     name='{}_link.json'.format(upload_item_element.name))
                if remote_name is None:
                    remote_name = link.name

                remote_filepath = '{}{}'.format(remote_path, remote_name)

                element = UploadElement(element_type='link',
                                        buffer=link,
                                        remote_filepath=remote_filepath,
                                        annotations_filepath=upload_annotations_element,
                                        item_metadata=item_metadata)
                elements.append(element)
            elif isinstance(upload_item_element, entities.Link):
                if remote_name is None:
                    remote_name = upload_item_element.name

                remote_filepath = '{}{}_link.json'.format(remote_path, remote_name)

                element = UploadElement(element_type='link',
                                        buffer=upload_item_element,
                                        remote_filepath=remote_filepath,
                                        annotations_filepath=upload_annotations_element,
                                        item_metadata=item_metadata)
                elements.append(element)
            elif isinstance(upload_item_element, entities.Similarity):
                if remote_name:
                    upload_item_element.name = remote_name
                if upload_item_element.name is None:
                    upload_item_element.name = '{}_similarity.json'.format(upload_item_element.ref)

                if not upload_item_element.name.endswith('.json'):
                    remote_filepath = '{}{}.json'.format(remote_path, upload_item_element.name)
                else:
                    remote_filepath = '{}{}'.format(remote_path, upload_item_element.name)

                element = UploadElement(element_type='collection',
                                        buffer=upload_item_element,
                                        remote_filepath=remote_filepath,
                                        annotations_filepath=upload_annotations_element,
                                        item_metadata=item_metadata)
                elements.append(element)
            elif isinstance(upload_item_element, entities.MultiView):
                if remote_name:
                    upload_item_element.name = remote_name

                if not upload_item_element.name.endswith('.json'):
                    upload_item_element.name = '{}.json'.format(upload_item_element.name)

                remote_filepath = '{}{}'.format(remote_path, upload_item_element.name)
                element = UploadElement(element_type='collection',
                                        buffer=upload_item_element,
                                        remote_filepath=remote_filepath,
                                        annotations_filepath=upload_annotations_element,
                                        item_metadata=item_metadata)
                elements.append(element)
            elif isinstance(upload_item_element, bytes) or \
                    isinstance(upload_item_element, io.BytesIO) or \
                    isinstance(upload_item_element, io.BufferedReader) or \
                    isinstance(upload_item_element, io.TextIOWrapper):
                # binary element
                if not hasattr(upload_item_element, "name") and remote_name is None:
                    raise PlatformException(
                        error="400",
                        message='Upload element type was bytes. Must put attribute "name" on buffer (with file name) '
                                'when uploading buffers or providing param "remote_name"')
                if remote_name is None:
                    remote_name = upload_item_element.name
                remote_filepath = remote_path + remote_name
                element = UploadElement(element_type='buffer',
                                        buffer=upload_item_element,
                                        remote_filepath=remote_filepath,
                                        annotations_filepath=upload_annotations_element,
                                        item_metadata=item_metadata)
                elements.append(element)
                # get size from binaries
                try:
                    total_size += upload_item_element.__sizeof__()
                except Exception:
                    logger.warning("Cant get binaries size")
            else:
                raise PlatformException(
                    error="400",
                    message="Unknown element type to upload ('local_path'). received type: {}. "
                            "known types (or list of those types): str (dir, file, url), bytes, io.BytesIO, "
                            "io.TextIOWrapper, Dataloop.Item, Dataloop.Link, "
                            "Dataloop.Similarity".format(type(upload_item_element)))
        return elements

    def _build_elements_from_df(self, df: pandas.DataFrame):
        elements = list()
        for index, row in df.iterrows():
            upload_item_element = row.get('local_path')
            remote_path = row.get('remote_path', '/')
            remote_name = row.get('remote_name', None)
            annotations_filepath = row.get('local_annotations_path', None)
            item_metadata = row.get('item_metadata', None)
            if not remote_path.endswith('/'):
                remote_path += '/'

            if os.path.isfile(upload_item_element):
                element_type = 'file'
                if remote_name is None:
                    remote_name = os.path.basename(upload_item_element)
            elif self.is_url(upload_item_element):
                element_type = 'url'
                if remote_name is None:
                    remote_name = str(upload_item_element.split('/')[-1])
            else:
                raise PlatformException("404", "Unknown local path: {}".format(upload_item_element))
            elements.append(UploadElement(element_type=element_type,
                                          buffer=upload_item_element,
                                          remote_filepath=remote_path + remote_name,
                                          annotations_filepath=annotations_filepath,
                                          item_metadata=item_metadata))

        return elements

    async def __single_async_upload(self,
                                    filepath,
                                    remote_path,
                                    uploaded_filename,
                                    last_try,
                                    mode,
                                    item_metadata,
                                    callback
                                    ):
        """
        Upload an item to dataset

        :param filepath: local filepath of the item
        :param remote_path: remote directory of filepath to upload
        :param uploaded_filename: optional - remote filename
        :param last_try: print log error only if last try
        :return: Item object
        """

        need_close = False
        if isinstance(filepath, str):
            # upload local file
            if not os.path.isfile(filepath):
                raise PlatformException(error="404", message="Filepath doesnt exists. file: {}".format(filepath))
            if uploaded_filename is None:
                uploaded_filename = os.path.basename(filepath)
            if os.path.isfile(filepath):
                item_type = 'file'
            else:
                item_type = 'dir'
            item_size = os.stat(filepath).st_size
            to_upload = open(filepath, 'rb')
            need_close = True

        else:
            # upload from buffer
            if isinstance(filepath, bytes):
                to_upload = io.BytesIO(filepath)
            elif isinstance(filepath, io.BytesIO):
                to_upload = filepath
            elif isinstance(filepath, io.BufferedReader):
                to_upload = filepath
            elif isinstance(filepath, io.TextIOWrapper):
                to_upload = filepath
            else:
                raise PlatformException("400", "Unknown input filepath type received: {}".format(type(filepath)))

            if uploaded_filename is None:
                if hasattr(filepath, "name"):
                    uploaded_filename = filepath.name
                else:
                    raise PlatformException(error="400",
                                            message="Must have filename when uploading bytes array (uploaded_filename)")

            item_size = to_upload.seek(0, 2)
            to_upload.seek(0)
            item_type = 'file'
        remote_url = "/datasets/{}/items".format(self.items_repository.dataset.id)
        try:
            response = await self.items_repository._client_api.upload_file_async(to_upload=to_upload,
                                                                                 item_type=item_type,
                                                                                 item_size=item_size,
                                                                                 item_metadata=item_metadata,
                                                                                 remote_url=remote_url,
                                                                                 uploaded_filename=uploaded_filename,
                                                                                 remote_path=remote_path,
                                                                                 callback=callback,
                                                                                 mode=mode)
        except Exception:
            raise
        finally:
            if need_close:
                to_upload.close()

        if response.ok:
            item = self.items_repository.items_entity.from_json(client_api=self.items_repository._client_api,
                                                                _json=response.json(),
                                                                dataset=self.items_repository.dataset)
        else:
            raise PlatformException(response)
        return item, response.headers.get('x-item-op', 'na')

    async def __upload_single_item_wrapper(self, i_item, element, pbar, reporter, mode):
        async with self.items_repository._client_api.event_loops('items.upload').semaphore('items.upload'):
            assert isinstance(element, UploadElement)
            item = False
            err = None
            trace = None
            saved_locally = False
            temp_dir = None
            action = 'na'
            remote_folder, remote_name = os.path.split(element.remote_filepath)

            if element.type == 'url':
                saved_locally, element.buffer, temp_dir = self.url_to_data(element.buffer)
            elif element.type == 'link':
                element.buffer = self.link(ref=element.buffer.ref, dataset_id=element.buffer.dataset_id,
                                           type=element.buffer.type, mimetype=element.buffer.mimetype)
            elif element.type == 'collection':
                element.buffer = element.buffer.to_bytes_io()

            for i_try in range(NUM_TRIES):
                try:
                    logger.debug("Upload item: {path}. Try {i}/{n}. Starting..".format(path=remote_name,
                                                                                       i=i_try + 1,
                                                                                       n=NUM_TRIES))
                    item, action = await self.__single_async_upload(filepath=element.buffer,
                                                                    mode=mode,
                                                                    item_metadata=element.item_metadata,
                                                                    remote_path=remote_folder,
                                                                    uploaded_filename=remote_name,
                                                                    last_try=(i_try + 1) == NUM_TRIES,
                                                                    callback=None)
                    logger.debug("Upload item: {path}. Try {i}/{n}. Success. Item id: {id}".format(path=remote_name,
                                                                                                   i=i_try + 1,
                                                                                                   n=NUM_TRIES,
                                                                                                   id=item.id))
                    if isinstance(item, entities.Item):
                        break
                    time.sleep(0.3 * (2 ** i_try))
                except Exception as e:
                    err = e
                    trace = traceback.format_exc()
                    logger.debug("Upload item: {path}. Try {i}/{n}. Fail.\n{trace}".format(path=remote_name,
                                                                                           i=i_try + 1,
                                                                                           n=NUM_TRIES,
                                                                                           trace=trace))

                finally:
                    if saved_locally and os.path.isdir(temp_dir):
                        shutil.rmtree(temp_dir)
            if item:
                if action in ['overwrite', 'created'] and element.annotations_filepath is not None:
                    try:
                        await self.__async_upload_annotations(annotations_filepath=element.annotations_filepath,
                                                              item=item)
                    except Exception:
                        logger.exception('Error uploading annotations to item id: {}'.format(item.id))

                reporter.set_index(i_item=i_item, status=action, output=item, success=True, ref=item.id)
            else:
                if isinstance(element.buffer, str):
                    ref = element.buffer
                elif hasattr(element.buffer, "name"):
                    ref = element.buffer.name
                else:
                    ref = 'Unknown'
                reporter.set_index(i_item=i_item, status='error', success=False,
                                   error="{}\n{}".format(err, trace), ref=ref)
            pbar.update()

    async def __async_upload_annotations(self, annotations_filepath, item):
        with open(annotations_filepath, 'r', encoding="utf8") as f:
            annotations = json.load(f)
        # wait for coroutines on the current event loop
        return await item.annotations._async_upload_annotations(annotations=annotations['annotations'])

    @staticmethod
    def url_to_data(url):
        chunk_size = 8192
        max_size = 30000000
        temp_dir = None

        # This will download the binaries from the URL user provided
        prepared_request = requests.Request(method='GET', url=url).prepare()
        with requests.Session() as s:
            retry = Retry(
                total=3,
                read=3,
                connect=3,
                backoff_factor=0.3,
            )
            adapter = HTTPAdapter(max_retries=retry)
            s.mount('http://', adapter)
            s.mount('https://', adapter)
            response = s.send(request=prepared_request, stream=True)

        total_length = response.headers.get("content-length")
        save_locally = int(total_length) > max_size

        if save_locally:
            # save to file
            temp_dir = tempfile.mkdtemp()
            temp_path = os.path.join(temp_dir, url.split('/')[-1])
            with open(temp_path, "wb") as f:
                for chunk in response.iter_content(chunk_size=chunk_size):
                    if chunk:  # filter out keep-alive new chunks
                        f.write(chunk)
            # save to output variable
            data = temp_path
        else:
            # save as byte stream
            data = io.BytesIO()
            for chunk in response.iter_content(chunk_size=chunk_size):
                if chunk:  # filter out keep-alive new chunks
                    data.write(chunk)
            # go back to the beginning of the stream
            data.seek(0)
            data.name = url.split('/')[-1]

        return save_locally, data, temp_dir

    @staticmethod
    def is_url(url):
        try:
            return validators.url(url)
        except Exception:
            return False

    @staticmethod
    def link(ref, type, mimetype=None, dataset_id=None):

        link_info = {'type': type,
                     'ref': ref}

        if mimetype:
            link_info['mimetype'] = mimetype

        if dataset_id is not None:
            link_info['datasetId'] = dataset_id

        _json = {'type': 'link',
                 'shebang': 'dataloop',
                 'metadata': {'dltype': 'link',
                              'linkInfo': link_info}}

        uploaded_byte_io = io.BytesIO()
        uploaded_byte_io.write(json.dumps(_json).encode())
        uploaded_byte_io.seek(0)

        return uploaded_byte_io



import numpy as np

from . import BaseAnnotationDefinition


class Box(BaseAnnotationDefinition):
    """
        Box annotation object
    """

    def __init__(self, left, top, right, bottom, label, attributes=None):
        self.type = "box"
        self.left = left
        self.top = top
        self.right = right
        self.bottom = bottom
        self.label = label
        if attributes is None:
            attributes = list()
        self.attributes = attributes

    @property
    def x(self):
        return [self.left, self.right]

    @property
    def y(self):
        return [self.top, self.bottom]

    @property
    def geo(self):
        return [
            [self.left, self.top],
            [self.right, self.bottom]
        ]

    def show(self, image, thickness, with_text, height, width, annotation_format, color):
        """
        Show annotation as ndarray
        :param image: empty or image to draw on
        :param thickness:
        :param with_text: not required
        :param height: item height
        :param width: item width
        :param annotation_format: options: dl.ViewAnnotationOptions.list()
        :param color: color
        :return: ndarray
        """
        try:
            import cv2
        except ImportError:
            self.logger.error(
                'Import Error! Cant import cv2. Annotations operations will be limited. import manually and fix errors')
            raise

        if thickness is None:
            thickness = 2

        # draw annotation
        image = cv2.rectangle(
            img=image,
            pt1=(int(np.round(self.left)), int(np.round(self.top))),
            pt2=(int(np.round(self.right)), int(np.round(self.bottom))),
            color=color,
            thickness=thickness,
            lineType=cv2.LINE_AA,
        )
        if with_text:
            image = self.add_text_to_image(image=image, annotation=self)
        return image

    def to_coordinates(self, color):
        return [{"x": float(x), "y": float(y), "z": 0} for x, y in self.geo]

    @staticmethod
    def from_coordinates(coordinates):
        return np.asarray([[pt["x"], pt["y"]] for pt in coordinates])

    @classmethod
    def from_json(cls, _json):
        if "coordinates" in _json:
            geo = cls.from_coordinates(_json["coordinates"])
        elif "data" in _json:
            geo = cls.from_coordinates(_json["data"])
        else:
            raise ValueError('can not find "coordinates" or "data" in annotation. id: {}'.format(_json["id"]))
        left = np.min(geo[:, 0])
        top = np.min(geo[:, 1])
        right = np.max(geo[:, 0])
        bottom = np.max(geo[:, 1])

        attributes = _json.get("attributes", list())

        return cls(
            left=left,
            top=top,
            right=right,
            bottom=bottom,
            label=_json["label"],
            attributes=attributes,
        )

    @classmethod
    def from_segmentation(cls, mask, label, attributes=None):
        """
        Convert binary mask to Polygon
        :param mask: binary mask (0,1)
        :param label: annotation label
        :param attributes: annotations list of attributes
        :return: Box annotation
        """
        left = 0
        if len(mask[1]) > 0:
            left = np.min(mask[1])

        top = 0
        if len(mask[0]) > 0:
            top = np.min(mask[0])

        right = 0
        if len(mask[1]) > 0:
            right = np.max(mask[1])

        bottom = 0
        if len(mask[0]) > 0:
            bottom = np.max(mask[0])

        return cls(left=left, top=top, right=right, bottom=bottom, label=label, attributes=attributes)



import importlib.util
import logging
import os
from shutil import copyfile

from .. import entities, repositories, exceptions, miscellaneous, assets, services

logger = logging.getLogger(name=__name__)


class Models:
    """
    Models Repository
    """

    def __init__(self, client_api: services.ApiClient, project: entities.Project = None):
        self._client_api = client_api
        self._project = project

    ############
    # entities #
    ############
    @property
    def project(self) -> entities.Project:
        if self._project is None:
            try:
                self._project = repositories.Projects(client_api=self._client_api).get()
            except exceptions.NotFound:
                raise exceptions.PlatformException(
                    error='2001',
                    message='Missing "project". need to set a Project entity or use project.models repository')
        return self._project

    @project.setter
    def project(self, project: entities.Project):
        if not isinstance(project, entities.Project):
            raise ValueError('Must input a valid Project entity')
        self._project = project

    ###########
    # methods #
    ###########
    def get(self, model_name=None, model_id=None, checkout=False, fetch=None) -> entities.Model:
        """
        Get model object

        :param checkout:
        :param model_id:
        :param model_name:
        :param fetch: optional - fetch entity from platform, default taken from cookie
        :return: model object
        """
        if fetch is None:
            fetch = self._client_api.fetch_entities

        if model_name is None and model_id is None:
            model = self.__get_from_cache()
            if model is None:
                raise exceptions.PlatformException(
                    error='400',
                    message='Checked out not found, must provide either model id or model name')
        elif fetch:
            if model_id is not None:
                success, response = self._client_api.gen_request(
                    req_type="get",
                    path="/models/{}".format(model_id))
                if not success:
                    raise exceptions.PlatformException(response)
                model = entities.Model.from_json(client_api=self._client_api,
                                                 _json=response.json(),
                                                 project=self._project)
            elif model_name is not None:
                models = self.list(entities.Filters(resource=entities.FiltersResource.MODEL,
                                                    field='name',
                                                    values=model_name))
                if models.items_count == 0:
                    raise exceptions.PlatformException(
                        error='404',
                        message='Model not found. Name: {}'.format(model_name))
                elif models.items_count > 1:
                    raise exceptions.PlatformException(
                        error='400',
                        message='More than one file found by the name of: {}'.format(model_name))
                model = models[0]
            else:
                raise exceptions.PlatformException(
                    error='400',
                    message='Checked out not found, must provide either model id or model name')
        else:
            model = entities.Model.from_json(_json={'id': model_id,
                                                    'name': model_name},
                                             client_api=self._client_api,
                                             project=self._project,
                                             is_fetched=False)

        if checkout:
            self.checkout(model=model)
        return model

    def _build_entities_from_response(self, response_items) -> miscellaneous.List[entities.Model]:
        jobs = [None for _ in range(len(response_items))]
        pool = self._client_api.thread_pools(pool_name='entity.create')

        # return triggers list
        for i_service, service in enumerate(response_items):
            jobs[i_service] = pool.apply_async(entities.Model._protected_from_json,
                                               kwds={'client_api': self._client_api,
                                                     '_json': service,
                                                     'project': self._project})
        # wait for all jobs
        _ = [j.wait() for j in jobs]
        # get all results
        results = [j.get() for j in jobs]
        # log errors
        _ = [logger.warning(r[1]) for r in results if r[0] is False]
        # return good jobs
        return miscellaneous.List([r[1] for r in results if r[0] is True])

    def _list(self, filters: entities.Filters):
        url = '/query/machine-learning'
        # request
        success, response = self._client_api.gen_request(req_type='POST',
                                                         path=url,
                                                         json_req=filters.prepare())
        if not success:
            raise exceptions.PlatformException(response)
        return response.json()

    def list(self, filters: entities.Filters = None) -> entities.PagedEntities:
        """
        List project models
        :return:
        """
        if filters is None:
            filters = entities.Filters(resource=entities.FiltersResource.MODEL)
            if self._project is not None:
                filters.add(field='projectId', values=self._project.id)

        # assert type filters
        if not isinstance(filters, entities.Filters):
            raise exceptions.PlatformException('400', 'Unknown filters type')

        paged = entities.PagedEntities(items_repository=self,
                                       filters=filters,
                                       page_offset=filters.page,
                                       page_size=filters.page_size,
                                       client_api=self._client_api)
        paged.get_page()
        return paged

    def build(self, model: entities.Model, local_path=None, from_local=None):
        """
        :param model: Model entity
        :param from_local: bool. use current directory to build
        :param local_path: local path of the model (if from_local=False - codebase will be downloaded)

        :return:
        """
        if from_local is None:
            from_local = False

        if from_local:
            # no need to download codebase
            if local_path is None:
                path = os.getcwd()
            else:
                path = local_path

        else:
            # download codebase locally
            if local_path is None:
                local_path = os.path.join(
                    services.service_defaults.DATALOOP_PATH,
                    "models",
                    model.name)

            codebase_id = model.codebase_id
            project = self._project if self._project is not None else model.project
            path = project.codebases.unpack(codebase_id=codebase_id, local_path=local_path)

        # load module from path
        entry_point = os.path.join(path, model.entry_point)
        spec = importlib.util.spec_from_file_location("AdapterModel", entry_point)
        foo = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(foo)
        adapter_model = foo.AdapterModel
        return adapter_model(model_entity=model)

    def push(self, model_name, description='', src_path=None, entry_point='adapter_model.py', codebase_id=None,
             checkout=False):
        """
        Push local model

        :param model_name: name of the model
        :param description: model description
        :param entry_point: location on the AdapterModel class
        :param codebase_id: codebase entity id. if none new will be created from src_path
        :param src_path: codebase location. if None pwd will be taken
        :param checkout: checkout entity to state
        :return:
        """
        # get project
        if self._project is None:
            raise exceptions.PlatformException('400', 'Repository does not have project. Please checkout a project,'
                                                      'or create model from a project models repository')

        # source path
        if src_path is None:
            if codebase_id is None:
                src_path = os.getcwd()
                logger.warning('No src_path is given, getting model information from cwd: {}'.format(src_path))

        # get or create codebase
        if codebase_id is None:
            codebase_id = self._project.codebases.pack(directory=src_path, name=model_name).id

        if not os.path.isfile(os.path.join(src_path, entry_point)):
            raise ValueError('entry point not found. filepath: {}'.format(os.path.join(src_path, entry_point)))
        # check if exist
        models = [model for model in self.list().all() if model.name == model_name]
        if len(models) > 0:
            model = self._create(codebase_id=codebase_id,
                                 model_name=model_name,
                                 description=description,
                                 entry_point=entry_point,
                                 push=True,
                                 model=models[0])
        else:
            model = self._create(codebase_id=codebase_id,
                                 entry_point=entry_point,
                                 description=description,
                                 model_name=model_name,
                                 push=False)
        if checkout:
            self.checkout(model=model)
        return model

    def _create(self,
                model_name,
                entry_point,
                description='',
                codebase_id=None,
                push=False,
                model=None) -> entities.Model:
        """
        Create a model in platform

        :param model:
        :param description: model description
        :param entry_point: location on the AdapterModel class
        :param codebase_id: optional - model codebase
        :param model_name: optional - default: 'default model'
        :param push:
        :return: Model Entity
        """
        if push:
            model.codebase_id = codebase_id
            return self.update(model=model)

        payload = {'name': model_name,
                   'codebaseId': codebase_id,
                   'entryPoint': entry_point,
                   'description': description}

        if self._project is not None:
            payload['projectId'] = self._project.id
        else:
            raise exceptions.PlatformException('400', 'Repository must have a project to perform this action')

        # request
        success, response = self._client_api.gen_request(req_type='post',
                                                         path='/models',
                                                         json_req=payload)

        # exception handling
        if not success:
            raise exceptions.PlatformException(response)

        # return entity
        return entities.Model.from_json(_json=response.json(),
                                        client_api=self._client_api,
                                        project=self._project)

    def delete(self, model: entities.Model = None, model_name=None, model_id=None):
        """
        Delete Model object

        :param model:
        :param model_name:
        :param model_id:
        :return: True
        """
        # get id and name
        if model_name is None or model_id is None:
            if model is None:
                model = self.get(model_id=model_id, model_name=model_name)
            model_id = model.id

        # request
        success, response = self._client_api.gen_request(
            req_type="delete",
            path="/models/{}".format(model_id)
        )

        # exception handling
        if not success:
            raise exceptions.PlatformException(response)

        # return results
        return True

    def update(self, model: entities.Model) -> entities.Model:
        """
        Update Model changes to platform

        :param model:
        :return: Model entity
        """
        # payload
        payload = model.to_json()

        # request
        success, response = self._client_api.gen_request(req_type='patch',
                                                         path='/models/{}'.format(model.id),
                                                         json_req=payload)

        # exception handling
        if not success:
            raise exceptions.PlatformException(response)

        # return entity
        return entities.Model.from_json(_json=response.json(),
                                        client_api=self._client_api,
                                        project=self._project)

    @staticmethod
    def generate(src_path=None):
        """
        Generate new model environment

        :return:
        """
        # src path
        if src_path is None:
            src_path = os.getcwd()
        if not os.path.isfile(os.path.join(src_path, '.gitignore')):
            copyfile(assets.paths.ASSETS_GITIGNORE_FILEPATH, os.path.join(src_path, '.gitignore'))
        copyfile(assets.paths.ASSETS_ADAPTER_MODEL_FILEPATH, os.path.join(src_path, assets.paths.ADAPTER_MODEL))
        logger.info('Successfully generated model')

    def __get_from_cache(self):
        model = self._client_api.state_io.get('model')
        if model is not None:
            model = entities.Model.from_json(_json=model, client_api=self._client_api, project=self._project)
        return model

    def checkout(self, model: entities.Model = None, model_id=None, model_name=None):
        """
        Checkout as model

        :param model_id:
        :param model:
        :param model_name:
        :return:
        """
        if model is None:
            model = self.get(model_id=model_id, model_name=model_name)
        self._client_api.state_io.put('model', model.to_json())
        logger.info("Checked out to model {}".format(model.name))



import traceback
import logging
import attr
import copy

from .. import entities

logger = logging.getLogger(name=__name__)


@attr.s
class Artifact(entities.Item):

    @staticmethod
    def _protected_from_json(_json, client_api, dataset=None, project=None, is_fetched=True):
        """
        Same as from_json but with try-except to catch if error
        :param _json:
        :param client_api:
        :param dataset:
        :return:
        """
        try:
            artifact = Artifact.from_json(_json=_json,
                                          client_api=client_api,
                                          dataset=dataset,
                                          project=None,
                                          is_fetched=True)
            status = True
        except Exception:
            artifact = traceback.format_exc()
            status = False
        return status, artifact

    @classmethod
    def from_json(cls, _json, client_api, dataset=None, project=None, is_fetched=True):
        """
        Build an Artifact entity object from a json
        :param project:
        :param _json: _json response from host
        :param dataset: Artifact's dataset
        :param client_api: client_api
        :param is_fetched: is Entity fetched from Platform
        :return: Artifact object
        """

        inst = cls(
            # sdk
            platform_dict=copy.deepcopy(_json),
            client_api=client_api,
            dataset=dataset,
            project=project,
            # params
            annotations_link=_json.get('annotations_link', None),
            annotations_count=_json.get('annotationsCount', None),
            createdAt=_json.get('createdAt', None),
            datasetId=_json.get('datasetId', None),
            thumbnail=_json.get('thumbnail', None),
            annotated=_json.get('annotated', None),
            dataset_url=_json.get('dataset', None),
            filename=_json.get('filename', None),
            metadata=_json.get('metadata', None),
            hidden=_json.get('hidden', False),
            stream=_json.get('stream', None),
            name=_json.get('name', None),
            type=_json.get('type', None),
            dir=_json.get('dir', None),
            url=_json.get('url', None),
            id=_json['id'])
        inst.is_fetched = is_fetched
        return inst



from . import BaseAnnotationDefinition


class Classification(BaseAnnotationDefinition):
    """
        Classification annotation object
    """

    def __init__(self, label, attributes=None):

        self.type = "class"
        self.label = label
        if attributes is None:
            attributes = list()
        self.attributes = attributes

    @property
    def x(self):
        return 0

    @property
    def y(self):
        return 0

    @property
    def geo(self):
        return list()

    @property
    def left(self):
        return 0

    @property
    def top(self):
        return 0

    @property
    def right(self):
        return 0

    @property
    def bottom(self):
        return 0

    def show(self, image, thickness, with_text, height, width, annotation_format, color):
        """
        Show annotation as ndarray
        :param image: empty or image to draw on
        :param thickness:
        :param with_text: not required
        :param height: item height
        :param width: item width
        :param annotation_format: options: dl.ViewAnnotationOptions.list()
        :param color: color
        :return: ndarray
        """
        if with_text:
            image = self.add_text_to_image(image=image, annotation=self)
        return image

    def to_coordinates(self, color):
        return list()

    @classmethod
    def from_json(cls, _json):
        attributes = _json.get("attributes", list())
        return cls(
            label=_json["label"],
            attributes=attributes,
        )



from behave import fixture
from behave import use_fixture
import os
import json
import logging


@fixture
def after_feature(context, feature):
    if hasattr(feature, 'bot'):
        try:
            feature.bot.delete()
        except Exception:
            logging.exception('Failed to delete bot')

    if hasattr(feature, 'dataloop_feature_project'):
        try:
            feature.dataloop_feature_project.delete(True, True)
        except Exception:
            logging.exception('Failed to delete project')

    # update api call json
    if hasattr(feature, 'dataloop_feature_dl'):
        try:
            api_calls_path = os.path.join(os.environ['DATALOOP_TEST_ASSETS'], 'api_calls.json')
            with open(api_calls_path, 'r') as f:
                api_calls = json.load(f)
            if context.feature.name in api_calls:
                api_calls[context.feature.name] += feature.dataloop_feature_dl.client_api.calls_counter.number
            else:
                api_calls[context.feature.name] = feature.dataloop_feature_dl.client_api.calls_counter.number
            with open(api_calls_path, 'w') as f:
                json.dump(api_calls, f)
        except Exception:
            logging.exception('Failed to update api calls')


@fixture
def after_tag(context, tag):
    if tag == 'services.delete':
        try:
            use_fixture(delete_services, context)
        except Exception:
            logging.exception('Failed to delete service')
    elif tag == 'packages.delete':
        try:
            use_fixture(delete_packages, context)
        except Exception:
            logging.exception('Failed to delete package')
    elif tag == 'bot.create':
        try:
            use_fixture(delete_bots, context)
        except Exception:
            logging.exception('Failed to delete bots')
    elif tag == 'second_project.delete':
        try:
            use_fixture(delete_second_project, context)
        except Exception:
            logging.exception('Failed to delete second project')
    elif tag == 'converter.platform_dataset.delete':
        try:
            use_fixture(delete_converter_dataset, context)
        except Exception:
            logging.exception('Failed to delete converter dataset')
    else:
        raise ValueError('unknown tag: {}'.format(tag))


@fixture
def delete_second_project(context):
    if hasattr(context, 'second_project'):
        context.second_project.delete(True, True)


@fixture
def delete_bots(context):
    if not hasattr(context, 'to_delete_projects_ids'):
        return

    all_deleted = True
    while context.to_delete_projects_ids:
        project_id = context.to_delete_projects_ids.pop(0)
        try:
            project = context.dl.projects.get(project_id=project_id)
            for bot in project.bots.list():
                try:
                    bot.delete()
                except:
                    logging.exception('Failed deleting bots: ')
                    all_deleted = False
                    pass
        except context.dl.exceptions.NotFound:
            pass
        except:
            logging.exception('Failed deleting bots: ')
    assert all_deleted


@fixture
def delete_packages(context):
    if not hasattr(context, 'to_delete_packages_ids'):
        return

    all_deleted = True
    while context.to_delete_packages_ids:
        package_id = context.to_delete_packages_ids.pop(0)
        try:
            context.dl.packages.delete(package_id=package_id)
        except context.dl.exceptions.NotFound:
            pass
        except:
            all_deleted = False
            logging.exception('Failed deleting package: ')
    assert all_deleted


@fixture
def delete_converter_dataset(context):
    if hasattr(context, 'platform_dataset'):
        context.platform_dataset.delete(True, True)


@fixture
def delete_services(context):
    if not hasattr(context, 'to_delete_services_ids'):
        return

    all_deleted = True
    while context.to_delete_services_ids:
        service_id = context.to_delete_services_ids.pop(0)
        try:
            context.dl.services.delete(service_id=service_id)
        except context.dl.exceptions.NotFound:
            pass
        except:
            all_deleted = False
            logging.exception('Failed deleting service: ')
    assert all_deleted



import multiprocessing
import threading
import traceback
import logging
import json
from copy import deepcopy

import requests
import tqdm
import os
import io

from PIL import Image
from requests.adapters import HTTPAdapter
from urllib3.util import Retry

from .. import entities, miscellaneous, PlatformException, services
from ..services import Reporter

logger = logging.getLogger(name=__name__)

NUM_TRIES = 3  # try to download 3 time before fail on item


class Downloader:
    def __init__(self, items_repository):
        self.items_repository = items_repository

    def download(self,
                 # filter options
                 filters: entities.Filters = None,
                 items=None,
                 # download options
                 local_path=None,
                 file_types=None,
                 save_locally=True,
                 num_workers=None,  # deprecated
                 overwrite=False,
                 annotation_options=None,
                 to_items_folder=True,
                 thickness=1,
                 with_text=False,
                 without_relative_path=None,
                 avoid_unnecessary_annotation_download=False
                 ):
        """
        Download dataset by filters.
        Filtering the dataset for items and save them local
        Optional - also download annotation, mask, instance and image mask of the item

        :param avoid_unnecessary_annotation_download:
        :param filters: Filters entity or a dictionary containing filters parameters
        :param items: download Item entity or item_id (or a list of item)
        :param overwrite: optional - default = False
        :param local_path: local folder or filename to save to.
        :param file_types: a list of file type to download. e.g ['video/webm', 'video/mp4', 'image/jpeg', 'image/png']
        :param num_workers: default - 32
        :param save_locally: bool. save to disk or return a buffer
        :param annotation_options: download annotations options. options: dl.ViewAnnotationOptions.list()
        :param to_items_folder: Create 'items' folder and download items to it
        :param with_text: optional - add text to annotations, default = False
        :param thickness: optional - line thickness, if -1 annotation will be filled, default =1
        :param without_relative_path: string - remote path - download items without the relative path from platform
        :return: Output (list)
        """

        ###################
        # Default options #
        ###################
        # annotation options
        if annotation_options is None:
            annotation_options = list()
        elif not isinstance(annotation_options, list):
            annotation_options = [annotation_options]
        for ann_option in annotation_options:
            if not isinstance(ann_option, entities.ViewAnnotationOptions):
                if ann_option not in entities.ViewAnnotationOptions.list():
                    raise PlatformException(
                        error='400',
                        message='Unknown annotation download option: {}, please chose from: {}'.format(
                            ann_option, entities.ViewAnnotationOptions.list()))
        if num_workers is not None:
            logger.warning('[DeprecationWarning] input argument "num_workers"'
                           ' will be deprecated from download() after version 1.17.0\n'
                           'To set number of processes use "dtlpy.client_api.num_processes=int()"')
            self.items_repository._client_api.num_processes = num_workers
        ##############
        # local path #
        ##############
        if local_path is None:
            # create default local path
            local_path = self.__default_local_path()

        if os.path.isdir(local_path):
            logger.info('Local folder already exists:{}. merge/overwrite according to "overwrite option"'.format(
                local_path))
        else:
            # check if filename
            _, ext = os.path.splitext(local_path)
            if not ext:
                path_to_create = local_path
                if local_path.endswith('*'):
                    path_to_create = os.path.dirname(local_path)
                logger.info("Creating new directory for download: {}".format(path_to_create))
                os.makedirs(path_to_create, exist_ok=True)

        #####################
        # items to download #
        #####################
        if items is not None:
            # convert input to a list
            if not isinstance(items, list):
                items = [items]
            # get items by id
            if isinstance(items[0], str):
                items = [self.items_repository.get(item_id=item_id) for item_id in items]
            elif isinstance(items[0], entities.Item):
                pass
            else:
                raise PlatformException(
                    error="400",
                    message='Unknown items type to download. Expecting str or Item entities. Got "{}" instead'.format(
                        type(items[0])
                    )
                )

            # convert to list of list (like pages and page)
            items_to_download = [items]
            num_items = len(items)
        else:
            # filters
            if filters is None:
                filters = entities.Filters()
            # file types
            if file_types is not None:
                filters.add(field='metadata.system.mimetype', values=file_types, operator='in')
            items_to_download = self.items_repository.list(filters=filters)
            num_items = items_to_download.items_count

        if num_items == 0:
            logger.warning('No items found! Nothing was downloaded')
            return list()
        ####################
        # annotations json #
        ####################
        # download annotations' json files in a new thread
        # items will start downloading and if json not exists yet - will download for each file
        if num_items > 1 and \
                annotation_options and \
                not avoid_unnecessary_annotation_download and \
                self.__need_to_download_annotations_zip(
                    items_count=self.__get_annotated_count(items=items, filters=filters)):
            # a new folder named 'json' will be created under the "local_path"
            logger.info("Downloading annotations formats: {}".format(annotation_options))
            thread = threading.Thread(target=self.download_annotations,
                                      kwargs={"dataset": self.items_repository.dataset,
                                              "local_path": local_path,
                                              'overwrite': overwrite},
                                      )
            thread.start()
        ###############
        # downloading #
        ###############
        # create result lists
        reporter = Reporter(num_workers=num_items, resource=Reporter.ITEMS_DOWNLOAD)
        jobs = [None for _ in range(num_items)]
        # pool
        pool = self.items_repository._client_api.thread_pools(pool_name='item.download')
        # download
        pbar = tqdm.tqdm(total=num_items, disable=self.items_repository._client_api.verbose.disable_progress_bar)
        try:
            i_item = 0
            for page in items_to_download:
                for item in page:
                    if item.type == "dir":
                        continue
                    if save_locally:
                        # get local file path
                        item_local_path, item_local_filepath = self.__get_local_filepath(
                            local_path=local_path,
                            without_relative_path=without_relative_path,
                            item=item,
                            to_items_folder=to_items_folder)

                        if os.path.isfile(item_local_filepath) and not overwrite:
                            logger.debug("File Exists: {}".format(item_local_filepath))
                            reporter.set_index(i_item=i_item, status='exist', output=item_local_filepath, success=True)
                            pbar.update()
                            if annotation_options and item.annotated:
                                # download annotations only
                                jobs[i_item] = pool.apply_async(
                                    self._download_img_annotations,
                                    kwds={
                                        "item": item,
                                        "img_filepath": item_local_filepath,
                                        "overwrite": overwrite,
                                        "annotation_options": annotation_options,
                                        "local_path": local_path,
                                        "thickness": thickness,
                                        "with_text": with_text
                                    },
                                )
                            i_item += 1
                            continue
                    else:
                        item_local_path = None
                        item_local_filepath = None

                    # download single item
                    jobs[i_item] = pool.apply_async(
                        self.__thread_download_wrapper,
                        kwds={
                            "i_item": i_item,
                            "item": item,
                            "item_local_path": item_local_path,
                            "item_local_filepath": item_local_filepath,
                            "save_locally": save_locally,
                            "annotation_options": annotation_options,
                            "reporter": reporter,
                            "pbar": pbar,
                            "overwrite": overwrite,
                            "thickness": thickness,
                            "with_text": with_text
                        },
                    )
                    i_item += 1
        except Exception:
            logger.exception('Error downloading:')
        finally:
            _ = [j.wait() for j in jobs if isinstance(j, multiprocessing.pool.ApplyResult)]
            pbar.close()
        # reporting
        n_download = reporter.status_count(status='download')
        n_exist = reporter.status_count(status='exist')
        n_error = reporter.status_count(status='error')
        logger.info("Number of files downloaded:{}".format(n_download))
        logger.info("Number of files exists: {}".format(n_exist))
        logger.info("Total number of files: {}".format(n_download + n_exist))

        # log error
        if n_error > 0:
            log_filepath = reporter.generate_log_files()
            logger.warning("Errors in {} files. See {} for full log".format(n_error, log_filepath))

        return reporter.output

    def __need_to_download_annotations_zip(self, items_count):
        try:
            if self.items_repository._dataset is None:
                if items_count > 1:
                    return True
                else:
                    return False
            else:
                return (items_count / self.items_repository._dataset.annotated) > 0.1
        except Exception:
            return False

    def __get_annotated_count(self, items, filters):
        try:
            copy_filters = deepcopy(filters)
            if items is not None:
                num_annotated = len([item for item in items if item.annotated])
            else:
                copy_filters.add(field='annotated', values=True)
                num_annotated = self.items_repository.list(filters=copy_filters).items_count
        except Exception:
            num_annotated = 0
        return num_annotated

    def __thread_download_wrapper(self, i_item,
                                  # item params
                                  item, item_local_path, item_local_filepath, save_locally, overwrite,
                                  # annotations params
                                  annotation_options, with_text, thickness,
                                  # threading params
                                  reporter, pbar):

        download = False
        err = None
        trace = None
        for i_try in range(NUM_TRIES):
            try:
                logger.debug("Download item: {path}. Try {i}/{n}. Starting..".format(path=item.filename,
                                                                                     i=i_try + 1,
                                                                                     n=NUM_TRIES))
                download = self.__thread_download(item=item,
                                                  save_locally=save_locally,
                                                  local_path=item_local_path,
                                                  local_filepath=item_local_filepath,
                                                  annotation_options=annotation_options,
                                                  overwrite=overwrite,
                                                  thickness=thickness,
                                                  with_text=with_text)
                logger.debug("Download item: {path}. Try {i}/{n}. Success. Item id: {id}".format(path=item.filename,
                                                                                                 i=i_try + 1,
                                                                                                 n=NUM_TRIES,
                                                                                                 id=item.id))
                if download:
                    break
            except Exception as e:
                logger.debug("Download item: {path}. Try {i}/{n}. Fail.".format(path=item.filename,
                                                                                i=i_try + 1,
                                                                                n=NUM_TRIES))
                err = e
                trace = traceback.format_exc()
        pbar.update()
        if not download:
            if err is None:
                err = self.items_repository._client_api.platform_exception
            reporter.set_index(i_item=i_item, status="error", ref=item.id, success=False,
                               error="{}\n{}".format(err, trace))
        else:
            reporter.set_index(i_item=i_item, status="download", output=download, success=True)

    def download_annotations(self, dataset: entities.Dataset, local_path, overwrite=False, remote_path=None):
        """
        Download annotations json for entire dataset

        :param remote_path:
        :param dataset: Dataset entity
        :param overwrite: optional - overwrite annotations if exist, default = false
        :param local_path:
        :return:
        """

        def download_single_chunk(w_filepath, w_url, remote_path=None):
            try:
                # remove heading of the url
                if remote_path is None:
                    w_url = w_url[w_url.find('/dataset'):]
                else:
                    w_url = '/datasets/{}/annotations/zip?directory={}'.format(dataset.id, remote_path)
                # get zip from platform
                success, response = dataset._client_api.gen_request(req_type="get",
                                                                    path=w_url,
                                                                    stream=True)

                if not success:
                    raise PlatformException(response)

                # downloading zip from platform
                with open(w_filepath, "wb") as f:
                    for chunk in response.iter_content(chunk_size=8192):
                        if chunk:  # filter out keep-alive new chunks
                            f.write(chunk)
                # unzipping annotations to directory
                miscellaneous.Zipping.unzip_directory(zip_filename=w_filepath,
                                                      to_directory=local_path)
            except Exception:
                raise PlatformException(error='400',
                                        message="Getting annotations from zip failed: {}".format(w_url))
            finally:
                # cleanup
                if os.path.isfile(w_filepath):
                    os.remove(w_filepath)

        local_path = os.path.join(local_path, "json")
        # only if json folder does not exist or exist and overwrite
        if not os.path.isdir(os.path.join(local_path, 'json')) or overwrite:
            # create local path to download and save to
            if not os.path.isdir(local_path):
                os.makedirs(local_path)
        if remote_path is None:
            urls = list()
            if isinstance(dataset.export['zip'], str):
                urls.append(dataset.export['zip'])
            elif isinstance(dataset.export['zip'], dict):
                for url in dataset.export['zip']['chunks']:
                    urls.append(url)
            pool = self.items_repository._client_api.thread_pools(pool_name='annotation.download')
            jobs = list()
            for i_url, url in enumerate(urls):
                # zip filepath
                zip_filepath = os.path.join(local_path, "annotations_{}.zip".format(i_url))
                # send url to pool
                jobs.append(pool.apply_async(download_single_chunk, kwds={'w_url': url,
                                                                          'w_filepath': zip_filepath}))
            _ = [j.wait() for j in jobs]
        else:
            zip_filepath = os.path.join(local_path, "annotations_{}.zip".format(remote_path.split('/')[-1]))
            download_single_chunk(w_url=None, w_filepath=zip_filepath, remote_path=remote_path)

    @staticmethod
    def _download_img_annotations(item: entities.Item,
                                  img_filepath,
                                  local_path,
                                  overwrite,
                                  annotation_options,
                                  thickness=1,
                                  with_text=False):

        # fix local path
        if local_path.endswith("/items") or local_path.endswith("\\items"):
            local_path = os.path.dirname(local_path)

        annotation_rel_path = item.filename[1:]

        # find annotations json
        annotations_json_filepath = os.path.join(local_path, "json", item.filename[1:])
        name, _ = os.path.splitext(annotations_json_filepath)
        annotations_json_filepath = name + ".json"

        if os.path.isfile(annotations_json_filepath):

            # if exists take from json file
            with open(annotations_json_filepath, "r", encoding="utf8") as f:
                data = json.load(f)
            if "annotations" in data:
                data = data["annotations"]
            annotations = entities.AnnotationCollection.from_json(_json=data, item=item)
        else:

            # if doesnt exist get from platform
            annotations = item.annotations.list()

        # get image shape
        if item.width is not None and item.height is not None:
            img_shape = (item.height, item.width)
        elif ('image' in item.mimetype and img_filepath is not None) or \
                ('json' in item.mimetype and img_filepath is not None):
            img_shape = Image.open(img_filepath).size[::-1]
        else:
            img_shape = (0, 0)

        # download all annotation options
        for option in annotation_options:
            # get path and create dirs
            annotation_filepath = os.path.join(local_path, option, annotation_rel_path)
            if not os.path.isdir(os.path.dirname(annotation_filepath)):
                os.makedirs(os.path.dirname(annotation_filepath), exist_ok=True)
            temp_path, ext = os.path.splitext(annotation_filepath)

            if option == entities.ViewAnnotationOptions.JSON:
                if not os.path.isfile(annotations_json_filepath):
                    annotations.download(
                        filepath=annotations_json_filepath,
                        annotation_format=option,
                        height=img_shape[0],
                        width=img_shape[1],
                    )
            elif option in [entities.ViewAnnotationOptions.MASK,
                            entities.ViewAnnotationOptions.INSTANCE,
                            entities.ViewAnnotationOptions.ANNOTATION_ON_IMAGE,
                            entities.ViewAnnotationOptions.VTT]:
                if option == entities.ViewAnnotationOptions.VTT:
                    annotation_filepath = temp_path + ".vtt"
                else:
                    annotation_filepath = temp_path + ".png"
                if not os.path.isfile(annotation_filepath) or overwrite:
                    # if not exists OR (exists AND overwrite)
                    if not os.path.exists(os.path.dirname(annotation_filepath)):
                        # create folder if not exists
                        os.makedirs(os.path.dirname(annotation_filepath), exist_ok=True)
                    if option == entities.ViewAnnotationOptions.ANNOTATION_ON_IMAGE and img_filepath is None:
                        raise PlatformException(
                            error="1002",
                            message="Missing image for annotation option dl.ViewAnnotationOptions.ANNOTATION_ON_IMAGE")
                    annotations.download(
                        filepath=annotation_filepath,
                        img_filepath=img_filepath,
                        annotation_format=option,
                        height=img_shape[0],
                        width=img_shape[1],
                        thickness=thickness,
                        with_text=with_text
                    )
            else:
                raise PlatformException(error="400", message="Unknown annotation option: {}".format(option))

    @staticmethod
    def __get_local_filepath(local_path, item, to_items_folder, without_relative_path=None):
        # create paths
        _, ext = os.path.splitext(local_path)
        if ext:
            # local_path is a filename
            local_filepath = local_path
            local_path = os.path.dirname(local_filepath)
        else:
            # if directory - get item's filename
            if to_items_folder:
                local_path = os.path.join(local_path, "items")
            if without_relative_path is not None:
                local_filepath = os.path.join(local_path, os.path.relpath(item.filename, without_relative_path))
            else:
                local_filepath = os.path.join(local_path, item.filename[1:])
        return local_path, local_filepath

    @staticmethod
    def __get_link_source(item):
        assert isinstance(item, entities.Item)
        if not item.filename.endswith('.json') or \
                'system' not in item.metadata or \
                'shebang' not in item.metadata['system'] or \
                item.metadata['system']['shebang']['dltype'] != 'link':
            return item, '', False

        # recursively get next id link item
        while item.filename.endswith('.json') and \
                'system' in item.metadata and \
                'shebang' in item.metadata['system'] and \
                'dltype' in item.metadata['system']['shebang'] and \
                item.metadata['system']['shebang']['dltype'] == 'link' and \
                'linkInfo' in item.metadata['system']['shebang'] and \
                item.metadata['system']['shebang']['linkInfo']['type'] == 'id':
            item = item.dataset.items.get(item_id=item.metadata['system']['shebang']['linkInfo']['ref'])

        # check if link
        if item.filename.endswith('.json') and \
                'system' in item.metadata and \
                'shebang' in item.metadata['system'] and \
                'dltype' in item.metadata['system']['shebang'] and \
                item.metadata['system']['shebang']['dltype'] == 'link' and \
                'linkInfo' in item.metadata['system']['shebang'] and \
                item.metadata['system']['shebang']['linkInfo']['type'] == 'url':
            url = item.metadata['system']['shebang']['linkInfo']['ref']
            return item, url, True
        else:
            return item, '', False

    def __thread_download(self,
                          item,
                          save_locally,
                          local_path,
                          local_filepath,
                          overwrite,
                          annotation_options,
                          chunk_size=8192,
                          thickness=1,
                          with_text=False
                          ):
        """
        Get a single item's binary data
        Calling this method will returns the item body itself , an image for example with the proper mimetype.

        :param item: Item entity to download
        :param save_locally: bool. save to file or return buffer
        :param local_path: optional - local folder or filename to save to.
        :param chunk_size: size of chunks to download - optional. default = 8192
        :param annotation_options: download annotations options: dl.ViewAnnotationOptions.list()
        :return:
        """
        # check if need to download image binary from platform
        need_to_download = True
        if save_locally and os.path.isfile(local_filepath):
            need_to_download = overwrite

        item, url, is_url = self.__get_link_source(item=item)

        response = None
        if need_to_download:
            if not is_url:
                result, response = self.items_repository._client_api.gen_request(req_type="get",
                                                                                 path="/items/{}/stream".format(
                                                                                     item.id),
                                                                                 stream=True)
                if not result:
                    raise PlatformException(response)
            else:
                response = self.get_url_stream(url=url)

        if need_to_download and save_locally:
            # save to file
            if not os.path.exists(os.path.dirname(local_filepath)):
                # create folder if not exists
                os.makedirs(os.path.dirname(local_filepath), exist_ok=True)

            # decide if create progress bar for item
            total_length = response.headers.get("content-length")
            one_file_pbar = None
            try:
                one_file_progress_bar = total_length is not None and int(total_length) > 10e6  # size larger than 10 MB
                if one_file_progress_bar:
                    one_file_pbar = tqdm.tqdm(total=int(total_length),
                                              unit='B',
                                              unit_scale=True,
                                              unit_divisor=1024,
                                              position=1,
                                              disable=self.items_repository._client_api.verbose.disable_progress_bar)
            except Exception as err:
                one_file_progress_bar = False
                logger.debug('Cant decide downloaded file length, bar will not be presented: {}'.format(err))

            # start download
            with open(local_filepath, "wb") as f:
                for chunk in response.iter_content(chunk_size=chunk_size):
                    if chunk:  # filter out keep-alive new chunks
                        f.write(chunk)
                        if one_file_progress_bar:
                            one_file_pbar.update(len(chunk))
            if one_file_progress_bar:
                one_file_pbar.close()
            # save to output variable
            data = local_filepath
            # if image - can download annotation mask
            if item.annotated and annotation_options:
                self._download_img_annotations(item=item,
                                               img_filepath=local_filepath,
                                               annotation_options=annotation_options,
                                               local_path=local_path,
                                               overwrite=overwrite,
                                               thickness=thickness,
                                               with_text=with_text
                                               )
        else:
            # save as byte stream
            data = io.BytesIO()
            for chunk in response.iter_content(chunk_size=chunk_size):
                if chunk:  # filter out keep-alive new chunks
                    data.write(chunk)
            # go back to the beginning of the stream
            data.seek(0)
            data.name = item.name
        return data

    def __default_local_path(self):

        # create default local path
        if self.items_repository._dataset is None:
            local_path = os.path.join(
                services.service_defaults.DATALOOP_PATH,
                "items",
            )
        else:
            if self.items_repository.dataset._project is None:
                # by dataset name
                local_path = os.path.join(
                    services.service_defaults.DATALOOP_PATH,
                    "datasets",
                    "{}_{}".format(self.items_repository.dataset.name, self.items_repository.dataset.id),
                )
            else:
                # by dataset and project name
                local_path = os.path.join(
                    services.service_defaults.DATALOOP_PATH,
                    "projects",
                    self.items_repository.dataset.project.name,
                    "datasets",
                    self.items_repository.dataset.name,
                )
        logger.info("Downloading to: {}".format(local_path))
        return local_path

    @staticmethod
    def get_url_stream(url):

        # This will download the binaries from the URL user provided
        prepared_request = requests.Request(method='GET', url=url).prepare()
        with requests.Session() as s:
            retry = Retry(
                total=3,
                read=3,
                connect=3,
                backoff_factor=0.3,
            )
            adapter = HTTPAdapter(max_retries=retry)
            s.mount('http://', adapter)
            s.mount('https://', adapter)
            response = s.send(request=prepared_request, stream=True)

        return response



import logging

from .. import entities, miscellaneous, PlatformException, exceptions, services

logger = logging.getLogger(name=__name__)


class Artifacts:
    """
    Artifacts repository
    """

    def __init__(self,
                 client_api: services.ApiClient,
                 project: entities.Project = None,
                 dataset: entities.Dataset = None):
        self._client_api = client_api
        if project is None and dataset is None:
            raise PlatformException('400', 'at least one must be not None: dataset or project')
        self._project = project
        self._dataset = dataset
        self._items_repository = None

    ############
    # entities #
    ############
    @property
    def project(self) -> entities.Project:
        assert isinstance(self._project, entities.Project)
        return self._project

    @property
    def dataset(self) -> entities.Dataset:
        if self._dataset is None:
            # get dataset from project
            try:
                self._dataset = self.project.datasets.get(dataset_name='Binaries')
            except exceptions.NotFound:
                self._dataset = None
            if self._dataset is None:
                logger.debug(
                    'Dataset for artifacts was not found. Creating... dataset name: "Binaries". project_id={id}'.format(
                        id=self.project.id))
                self._dataset = self.project.datasets.create(dataset_name='Binaries')
                # add system to metadata
                if 'metadata' not in self._dataset.to_json():
                    self._dataset.metadata = dict()
                if 'system' not in self._dataset.metadata:
                    self._dataset.metadata['system'] = dict()
                self._dataset.metadata['system']['scope'] = 'system'
                self.project.datasets.update(dataset=self._dataset, system_metadata=True)
        return self._dataset

    ################
    # repositories #
    ################
    @property
    def items_repository(self):
        if self._items_repository is None:
            # load Binaries dataset
            # load items repository
            self._items_repository = self.dataset.items
            self._items_repository.set_items_entity(entities.Artifact)
        return self._items_repository

    ###########
    # methods #
    ###########
    @staticmethod
    def _build_path_header(package_name=None, package=None, execution_id=None, execution=None, model_name=None,
                           snapshot_name=None):
        remote_path = '/artifacts'
        if package_name is not None or package is not None:
            if package is not None:
                package_name = package.name
            remote_path += '/packages/{}'.format(package_name)
        if execution_id is not None or execution is not None:
            if execution is not None:
                execution_id = execution.id
            remote_path += '/executions/{}'.format(execution_id)
        if model_name is not None:
            remote_path += '/models/{}'.format(model_name)
        if snapshot_name is not None:
            remote_path += '/snapshots/{}'.format(snapshot_name)

        return remote_path

    def list(self, execution_id=None, package_name=None) -> miscellaneous.List[entities.Artifact]:
        """
        List of artifacts
        :return: list of artifacts
        """
        filters = entities.Filters()
        remote_path = self._build_path_header(package_name=package_name, execution_id=execution_id)
        remote_path += '/*'
        filters.add(field='filename', values=remote_path)
        pages = self.items_repository.list(filters=filters)
        items = [item for page in pages for item in page]
        return miscellaneous.List(items)

    def get(self, artifact_id=None, artifact_name=None,
            execution_id=None, package_name=None) -> entities.Artifact:
        """

        Get an artifact object by name, id or type
        If by name or type - need to input also execution/task id for the artifact folder
        :param package_name:
        :param artifact_id: optional - search by id
        :param artifact_name:
        :param execution_id:
        :return: Artifact object
        """
        if artifact_id is not None:
            artifact = self.items_repository.get(item_id=artifact_id)
            return artifact
        elif artifact_name is not None:
            artifacts = self.list(execution_id=execution_id, package_name=package_name)
            artifact = [artifact for artifact in artifacts if artifact.name == artifact_name]
            if len(artifact) == 1:
                artifact = artifact[0]
            elif len(artifact) > 1:
                raise PlatformException('404', 'More Than one Artifact found')
            else:
                raise PlatformException('404', 'Artifact not found')
            return artifact
        else:
            msg = 'one input must be not None: artifact_id or artifact_name'
            raise ValueError(msg)

    def download(self, artifact_id=None, artifact_name=None,
                 execution_id=None, package_name=None,
                 local_path=None, overwrite=False, save_locally=True):
        """

        Download artifact binary.
        Get artifact by name, id or type

        :param save_locally:
        :param overwrite: optional - default = False
        :param artifact_id: optional - search by id
        :param local_path: artifact will be saved to this filepath
        :param artifact_name:
        :param execution_id:
        :param package_name:
        :return: Artifact object
        """
        if artifact_id is not None:
            artifact = self.items_repository.download(items=artifact_id,
                                                      save_locally=save_locally,
                                                      local_path=local_path,
                                                      overwrite=overwrite)
        elif artifact_name is None:
            if all(elem is None for elem in [package_name, execution_id]):
                raise PlatformException(error='400', message='Must input package or execution (id or entity)')

            remote_path = self._build_path_header(package_name=package_name,
                                                  execution_id=execution_id)
            without_relative_path = remote_path
            remote_path += '/*'
            filters = entities.Filters()
            filters.add(field='filename', values=remote_path)
            artifact = self.items_repository.download(filters=filters,
                                                      save_locally=save_locally,
                                                      local_path=local_path,
                                                      to_items_folder=False,
                                                      overwrite=overwrite,
                                                      without_relative_path=without_relative_path)
        else:
            artifact_obj = self.get(artifact_id=artifact_id,
                                    execution_id=execution_id,
                                    package_name=package_name,
                                    artifact_name=artifact_name)

            artifact = artifact_obj.download(save_locally=save_locally,
                                             local_path=local_path,
                                             overwrite=overwrite)
        return artifact

    def upload(self,
               # what to upload
               filepath,
               # where to upload
               package_name=None, package=None, execution_id=None, execution=None, model_name=None,
               snapshot_name=None,
               # add information
               overwrite=False):
        """

        Upload binary file to artifact. get by name, id or type.
        If artifact exists - overwriting binary
        Else and if create==True a new artifact will be created and uploaded

        :param overwrite: optional - default = False
        :param filepath: local binary file
        :param model_name:
        :param snapshot_name:
        :param package_name:
        :param package:
        :param execution_id:
        :param execution:
        :return: Artifact Object
        """
        remote_path = self._build_path_header(package_name=package_name,
                                              package=package,
                                              execution=execution,
                                              execution_id=execution_id,
                                              model_name=model_name,
                                              snapshot_name=snapshot_name)

        if all(elem is None for elem in [package_name, package, execution_id, execution, model_name, snapshot_name]):
            raise ValueError('Must input package or execution (id or entity)')

        artifact = self.items_repository.upload(local_path=filepath,
                                                remote_path=remote_path,
                                                overwrite=overwrite)

        logger.debug('Artifact uploaded successfully')
        return artifact

    def delete(self, artifact_id=None, artifact_name=None, execution_id=None, package_name=None):
        if artifact_id is not None or artifact_name is not None:
            artifacts = [self.get(artifact_id=artifact_id, artifact_name=artifact_name)]
        elif execution_id is not None or package_name is not None:
            artifacts = self.list(execution_id=execution_id, package_name=package_name)
        else:
            raise PlatformException('400',
                                    'Must provide one of: artifact_id, artifact_name, execution_id, package_name')

        # TODO - Temporary
        for artifact in artifacts:
            self.items_repository.delete(item_id=artifact.id)

        # TODO - when delete with filters is enabled again use this:
        # values = [artifact.id for artifact in artifacts]
        # self.items_repository.delete(filters=entities.Filters(field='id', values=values, operator='in'))

        return True



def get_callback(progress):
    from keras.callbacks import Callback
    import numpy as np
    import logging
    import time

    logger = logging.getLogger(name=__name__)

    class PiperProgressReporter(Callback):
        def __init__(self, progress):
            super(PiperProgressReporter, self).__init__()
            self.progress = progress
            self.results = dict()
            self.epoch_time_start = None

        def on_train_begin(self, logs=None):
            self.results = dict()

        def on_epoch_begin(self, batch, logs=None):
            self.epoch_time_start = time.time()

        def on_epoch_end(self, epoch, logs=None):
            logs_dict = dict(zip(list(logs.keys()), [float(num) for num in np.array(list(logs.values())).tolist()]))
            self.results[epoch] = logs_dict
            self.results[epoch]['runtime'] = time.time() - self.epoch_time_start
            self.progress.report_output(output={'epoch': epoch,
                                                'logs': logs_dict})

    return PiperProgressReporter(progress)



"""
Datasets Repository
"""

import os
import tqdm
import logging
from urllib.parse import urlencode
from multiprocessing.pool import ThreadPool

from .. import entities, repositories, miscellaneous, exceptions, services

logger = logging.getLogger(name=__name__)


class Datasets:
    """
    Datasets repository
    """

    def __init__(self, client_api: services.ApiClient, project: entities.Project = None):
        self._client_api = client_api
        self._project = project

    ############
    # entities #
    ############
    @property
    def project(self) -> entities.Project:
        if self._project is None:
            # try get checkout
            project = self._client_api.state_io.get('project')
            if project is not None:
                self._project = entities.Project.from_json(_json=project, client_api=self._client_api)
        if self._project is None:
            raise exceptions.PlatformException(
                error='2001',
                message='Cannot perform action WITHOUT Project entity in Datasets repository.'
                        ' Please checkout or set a project')
        assert isinstance(self._project, entities.Project)
        return self._project

    @project.setter
    def project(self, project: entities.Project):
        if not isinstance(project, entities.Project):
            raise ValueError('Must input a valid Project entity')
        self._project = project

    ###########
    # methods #
    ###########
    def __get_from_cache(self) -> entities.Dataset:
        dataset = self._client_api.state_io.get('dataset')
        if dataset is not None:
            dataset = entities.Dataset.from_json(_json=dataset,
                                                 client_api=self._client_api,
                                                 datasets=self,
                                                 project=self._project)
        return dataset

    def __get_by_id(self, dataset_id) -> entities.Dataset:
        success, response = self._client_api.gen_request(req_type='get',
                                                         path='/datasets/{}'.format(dataset_id))
        if dataset_id is None or dataset_id == '':
            raise exceptions.PlatformException('400', 'Please checkout a dataset')

        if success:
            dataset = entities.Dataset.from_json(client_api=self._client_api,
                                                 _json=response.json(),
                                                 datasets=self,
                                                 project=self._project)
        else:
            raise exceptions.PlatformException(response)
        return dataset

    def __get_by_identifier(self, identifier=None) -> entities.Dataset:
        datasets = self.list()
        datasets_by_name = [dataset for dataset in datasets if identifier in dataset.name or identifier in dataset.id]
        if len(datasets_by_name) == 1:
            return datasets_by_name[0]
        elif len(datasets_by_name) > 1:
            raise Exception('Multiple datasets with this name exist')
        else:
            raise Exception("Dataset not found")

    def open_in_web(self, dataset_name=None, dataset_id=None, dataset=None):
        if dataset is None:
            dataset = self.get(dataset_id=dataset_id, dataset_name=dataset_name)
        self._client_api._open_in_web(resource_type='dataset',
                                      project_id=dataset.project.id, # need to get the project otherwise will fail
                                      dataset_id=dataset.id)

    def checkout(self, identifier=None, dataset_name=None, dataset_id=None, dataset=None):
        """
        Check-out a project
        :param dataset:
        :param dataset_id:
        :param dataset_name:
        :param identifier: project name or partial id
        :return:
        """
        if dataset is None:
            if dataset_id is not None or dataset_name is not None:
                dataset = self.get(dataset_id=dataset_id, dataset_name=dataset_name)
            elif identifier is not None:
                dataset = self.__get_by_identifier(identifier=identifier)
            else:
                raise exceptions.PlatformException(error='400',
                                                   message='Must provide partial/full id/name to checkout')
        self._client_api.state_io.put('dataset', dataset.to_json())
        logger.info('Checked out to dataset {}'.format(dataset.name))

    def list(self, name=None, creator=None) -> miscellaneous.List[entities.Dataset]:
        """
        List all datasets.

        :return: List of datasets
        """
        url = '/datasets'

        query_params = {
            'name': name,
            'creator': creator
        }

        if self._project is not None:
            query_params['projects'] = self.project.id

        url += '?{}'.format(urlencode({key: val for key, val in query_params.items() if val is not None}, doseq=True))

        success, response = self._client_api.gen_request(req_type='get',
                                                         path=url)
        if success:
            pool = self._client_api.thread_pools('entity.create')
            datasets_json = response.json()
            jobs = [None for _ in range(len(datasets_json))]
            # return triggers list
            for i_dataset, dataset in enumerate(datasets_json):
                jobs[i_dataset] = pool.apply_async(entities.Dataset._protected_from_json,
                                                   kwds={'client_api': self._client_api,
                                                         '_json': dataset,
                                                         'datasets': self,
                                                         'project': self.project})
            # wait for all jobs
            _ = [j.wait() for j in jobs]
            # get all results
            results = [j.get() for j in jobs]
            # log errors
            _ = [logger.warning(r[1]) for r in results if r[0] is False]
            # return good jobs
            datasets = miscellaneous.List([r[1] for r in results if r[0] is True])
        else:
            raise exceptions.PlatformException(response)
        return datasets

    def get(self, dataset_name=None, dataset_id=None, checkout=False, fetch=None) -> entities.Dataset:
        """
        Get dataset by name or id

        :param checkout:
        :param dataset_name: optional - search by name
        :param dataset_id: optional - search by id
        :param fetch: optional - fetch entity from platform, default taken from cookie
        :return: Dataset object
        """
        if fetch is None:
            fetch = self._client_api.fetch_entities

        if dataset_id is None and dataset_name is None:
            dataset = self.__get_from_cache()
            if dataset is None:
                raise exceptions.PlatformException(
                    error='400',
                    message='Checked out not found, must provide either dataset id or dataset name')
        elif fetch:
            if dataset_id is not None and dataset_id != '':
                dataset = self.__get_by_id(dataset_id)
            elif dataset_name is not None:
                datasets = self.list(name=dataset_name)
                if not datasets:
                    # empty list
                    raise exceptions.PlatformException('404', 'Dataset not found. Name: {}'.format(dataset_name))
                    # dataset = None
                elif len(datasets) > 1:
                    raise exceptions.PlatformException('400', 'More than one dataset with same name.')
                else:
                    dataset = datasets[0]
            else:
                raise exceptions.PlatformException(
                    error='404',
                    message='No input and no checked-out found')
        else:
            dataset = entities.Dataset.from_json(_json={'id': dataset_id,
                                                        'name': dataset_id},
                                                 client_api=self._client_api,
                                                 datasets=self,
                                                 project=self._project,
                                                 is_fetched=False)
        assert isinstance(dataset, entities.Dataset)
        if checkout:
            self.checkout(dataset=dataset)
        return dataset

    def delete(self, dataset_name=None, dataset_id=None, sure=False, really=False):
        """
        Delete a dataset forever!
        :param dataset_name: optional - search by name
        :param dataset_id: optional - search by id
        :param sure: are you sure you want to delete?
        :param really: really really?

        :return: True
        """
        if sure and really:
            dataset = self.get(dataset_name=dataset_name, dataset_id=dataset_id)
            success, response = self._client_api.gen_request(req_type='delete',
                                                             path='/datasets/{}'.format(dataset.id))
            if not success:
                raise exceptions.PlatformException(response)
            logger.info('Dataset {} was deleted successfully'.format(dataset.name))
            return True
        else:
            raise exceptions.PlatformException(
                error='403',
                message='Cant delete dataset from SDK. Please login to platform to delete')

    def update(self, dataset: entities.Dataset, system_metadata=False) -> entities.Dataset:
        """
        Update dataset field
        :param dataset: Dataset entity
        :param system_metadata: bool
        :return: Dataset object
        """
        url_path = '/datasets/{}'.format(dataset.id)
        if system_metadata:
            url_path += '?system=true'
        success, response = self._client_api.gen_request(req_type='patch',
                                                         path=url_path,
                                                         json_req=dataset.to_json())
        if success:
            logger.info('Dataset was updated successfully')
            return dataset
        else:
            raise exceptions.PlatformException(response)

    def directory_tree(self, dataset: entities.Dataset = None, dataset_name=None, dataset_id=None):
        """
        Get dataset's directory tree
        :return:
        """
        if dataset is None and dataset_name is None and dataset_id is None:
            raise exceptions.PlatformException('400', 'Must provide dataset, dataset name or dataset id')
        if dataset_id is None:
            if dataset is None:
                dataset = self.get(dataset_name=dataset_name)
            dataset_id = dataset.id

        url_path = '/datasets/{}/directoryTree'.format(dataset_id)

        success, response = self._client_api.gen_request(req_type='get',
                                                         path=url_path)

        if success:
            return entities.DirectoryTree(_json=response.json())
        else:
            raise exceptions.PlatformException(response)

    def clone(self, dataset_id, clone_name, filters=None, with_items_annotations=True, with_metadata=True,
              with_task_annotations_status=True):
        """
        Clone a dataset

        :param dataset_id: to clone dataset
        :param clone_name: new dataset name
        :param filters: Filters entity or a query dict
        :param with_items_annotations:
        :param with_metadata:
        :param with_task_annotations_status:
        :return:
        """
        if filters is None:
            filters = entities.Filters().prepare()
        elif isinstance(filters, entities.Filters):
            filters = filters.prepare()
        else:
            raise exceptions.PlatformException(
                error='400',
                message='"filters" must be a dl.Filters entity. got: {}'.format(type(filters)))

        payload = {
            "name": clone_name,
            "filter": filters,
            "cloneDatasetParams": {
                "withItemsAnnotations": with_items_annotations,
                "withMetadata": with_metadata,
                "withTaskAnnotationsStatus": with_task_annotations_status
            }
        }
        success, response = self._client_api.gen_request(req_type='post',
                                                         path='/datasets/{}/clone'.format(dataset_id),
                                                         json_req=payload)

        if success:
            return entities.Dataset.from_json(_json=response.json(),
                                              project=self._project,
                                              datasets=self,
                                              client_api=self._client_api)
        else:
            raise exceptions.PlatformException(response)

    def merge(self, merge_name, dataset_ids, project_ids, with_items_annotations=True, with_metadata=True,
              with_task_annotations_status=True):
        payload = {
            "name": merge_name,
            "datasetsIds": dataset_ids,
            "projectIds": project_ids,
            "mergeDatasetParams": {
                "withItemsAnnotations": with_items_annotations,
                "withMetadata": with_metadata,
                "withTaskAnnotationsStatus": with_task_annotations_status
            }
        }
        success, response = self._client_api.gen_request(req_type='post',
                                                         path='/datasets/merge',
                                                         json_req=payload)

        if success:
            return entities.Dataset.from_json(_json=response.json(),
                                              project=self._project,
                                              datasets=self,
                                              client_api=self._client_api)
        else:
            raise exceptions.PlatformException(response)

    def create(self, dataset_name, labels=None, driver=None, attributes=None, ontology_ids=None,
               checkout=False) -> entities.Dataset:
        """
        Create a new dataset

        :param checkout:
        :param ontology_ids: optional - dataset ontology
        :param dataset_name: name
        :param attributes: dataset's ontology's attributes
        :param labels: dictionary of {tag: color} or list of label entities
        :param driver: optional
        :return: Dataset object
        """
        # labels to list
        if labels is not None:
            if not all(isinstance(label, entities.Label) for label in labels):
                labels = entities.Dataset.serialize_labels(labels)
        else:
            labels = list()
        # get creator from token
        payload = {'name': dataset_name,
                   'projects': [self.project.id]}
        if driver is not None:
            payload['driver'] = driver
        success, response = self._client_api.gen_request(req_type='post',
                                                         path='/datasets',
                                                         json_req=payload)
        if success:
            dataset = entities.Dataset.from_json(client_api=self._client_api,
                                                 _json=response.json(),
                                                 datasets=self,
                                                 project=self.project)
            # create ontology and recipe
            dataset = dataset.recipes.create(ontology_ids=ontology_ids, labels=labels, attributes=attributes).dataset
            # # patch recipe to dataset
            # dataset = self.update(dataset=dataset, system_metadata=True)
        else:
            raise exceptions.PlatformException(response)
        logger.info('Dataset was created successfully. Dataset id: {}'.format(dataset.id))
        assert isinstance(dataset, entities.Dataset)
        if checkout:
            self.checkout(dataset=dataset)
        return dataset

    @staticmethod
    def download_annotations(dataset,
                             local_path=None,
                             filters=None,
                             annotation_options=None,
                             overwrite=False,
                             thickness=1,
                             with_text=False,
                             num_workers=32,
                             remote_path=None):
        def download_single(i_item, i_img_filepath, i_local_path, i_overwrite, i_annotation_options,
                            i_thickness, i_with_text):
            try:
                repositories.Downloader._download_img_annotations(item=i_item,
                                                                  img_filepath=i_img_filepath,
                                                                  local_path=i_local_path,
                                                                  overwrite=i_overwrite,
                                                                  annotation_options=i_annotation_options,
                                                                  thickness=i_thickness,
                                                                  with_text=i_with_text)
            except Exception:
                logger.error('Failed to download annotation for item: {}'.format(item.name))

            progress.update(1)

        if local_path is None:
            if dataset.project is None:
                # by dataset name
                local_path = os.path.join(
                    services.service_defaults.DATALOOP_PATH,
                    "datasets",
                    "{}_{}".format(dataset.name, dataset.id),
                )
            else:
                # by dataset and project name
                local_path = os.path.join(
                    services.service_defaults.DATALOOP_PATH,
                    "projects",
                    dataset.project.name,
                    "datasets",
                    dataset.name,
                )

        downloader = repositories.Downloader(items_repository=dataset.items)

        # check if need to only download zip
        if filters is None:
            filters = entities.Filters()
            if annotation_options is None:
                downloader.download_annotations(dataset=dataset,
                                                local_path=local_path,
                                                overwrite=overwrite,
                                                remote_path=remote_path)
                return local_path

        filters.add(field='annotated', values=True)

        if annotation_options is None:
            annotation_options = [entities.ViewAnnotationOptions.JSON]
        if not isinstance(annotation_options, list):
            annotation_options = [annotation_options]

        pages = dataset.items.list(filters=filters)

        if pages.items_count > dataset.annotated / 10:
            downloader.download_annotations(dataset=dataset,
                                            local_path=local_path,
                                            overwrite=overwrite,
                                            remote_path=remote_path)

        pool = ThreadPool(processes=num_workers)
        progress = tqdm.tqdm(total=pages.items_count)
        for page in pages:
            for item in page:
                pool.apply_async(
                    func=download_single,
                    kwds={
                        'i_item': item,
                        'i_img_filepath': None,
                        'i_local_path': local_path,
                        'i_overwrite': overwrite,
                        'i_annotation_options': annotation_options,
                        'i_thickness': thickness,
                        'i_with_text': with_text
                    }
                )

        pool.close()
        pool.join()
        pool.terminate()
        progress.close()

        return local_path



def main():
    from keras.callbacks import Callback
    import dtlpy as dl
    import numpy as np
    import logging
    import time
    import json
    import os

    logger = logging.getLogger(name=__name__)

    class ProgressViewer(Callback):
        def __init__(self, session_id, directory=None):
            super(ProgressViewer, self).__init__()
            # init Dataloop instance
            # get sessions artifact
            self.session = dl.sessions.get(session_id=session_id)
            artifacts = self.session.artifacts.list()
            self.artifact = None
            for artifact in artifacts:
                if artifact.type == 'progress':
                    self.artifact = artifact
                    logger.info('Progress artifact found. overwriting. artifact_id: %s' % self.artifact.id)
                    break
            if self.artifact is None:
                self.artifact = self.session.artifacts.create(artifact_name='progress.yml',
                                                              artifact_type='progress',
                                                              description='update progress on each epoch')

                logger.info('[INFO] Creating progress artifact. artifact_id: %s' % self.artifact.id)
            if directory is None:
                directory = './results'
            if not os.path.isdir(directory):
                os.makedirs(directory)
            self.filename = os.path.join(directory, 'progress.yml')
            self.results = dict()
            self.epoch_time_start = None

        def on_train_begin(self, logs=None):
            self.results = dict()

        def on_epoch_begin(self, batch, logs=None):
            self.epoch_time_start = time.time()

        def on_epoch_end(self, epoch, logs=None):
            self.results[epoch] = dict(zip(list(logs.keys()), np.array(list(logs.values())).tolist()))
            self.results[epoch]['runtime'] = time.time() - self.epoch_time_start
            with open(self.filename, 'w') as f:
                json.dump(self.results, f)
            self.session.artifacts.upload(filepath=self.filename,
                                          artifact_name='progress.yml',
                                          artifact_type='progress')

    return ProgressViewer



from collections import namedtuple
import traceback
import logging
import attr
import os

from .. import repositories, entities, services

logger = logging.getLogger(name=__name__)


@attr.s
class Dataset(entities.BaseEntity):
    """
    Dataset object
    """
    # dataset information
    id = attr.ib()
    url = attr.ib()
    name = attr.ib()
    annotated = attr.ib(repr=False)
    creator = attr.ib()
    projects = attr.ib(repr=False)
    itemsCount = attr.ib()
    metadata = attr.ib(repr=False)
    directoryTree = attr.ib(repr=False)
    export = attr.ib(repr=False)

    # name change when to_json
    created_at = attr.ib()
    items_url = attr.ib(repr=False)
    readable_type = attr.ib(repr=False)
    access_level = attr.ib(repr=False)
    driver = attr.ib(repr=False)

    # api
    _client_api = attr.ib(type=services.ApiClient, repr=False)
    _instance_map = attr.ib(default=None, repr=False)

    # entities
    _project = attr.ib(default=None, repr=False)

    # repositories
    _datasets = attr.ib(repr=False, default=None)
    _repositories = attr.ib(repr=False)

    # defaults
    _ontology_ids = attr.ib(default=None, repr=False)
    _labels = attr.ib(default=None, repr=False)
    _directory_tree = attr.ib(default=None, repr=False)

    @staticmethod
    def _protected_from_json(project: entities.Project,
                             _json: dict,
                             client_api: services.ApiClient,
                             datasets=None,
                             is_fetched=True):
        """
        Same as from_json but with try-except to catch if error
         :param is_fetched: is Entity fetched from Platform
        :param _json: _json response from host
        :param project: dataset's project
        :param datasets: Datasets repository
        :param client_api: client_api
        :return: Dataset object
        """
        try:
            dataset = Dataset.from_json(project=project,
                                        _json=_json,
                                        client_api=client_api,
                                        datasets=datasets,
                                        is_fetched=is_fetched)
            status = True
        except Exception:
            dataset = traceback.format_exc()
            status = False
        return status, dataset

    @classmethod
    def from_json(cls,
                  project: entities.Project,
                  _json: dict,
                  client_api: services.ApiClient,
                  datasets=None,
                  is_fetched=True):
        """
        Build a Dataset entity object from a json

        :param is_fetched: is Entity fetched from Platform
        :param _json: _json response from host
        :param project: dataset's project
        :param datasets: Datasets repository
        :param client_api: client_api
        :return: Dataset object
        """
        inst = cls(metadata=_json.get('metadata', None),
                   directoryTree=_json.get('directoryTree', None),
                   readable_type=_json.get('readableType', None),
                   access_level=_json.get('accessLevel', None),
                   created_at=_json.get('createdAt', None),
                   itemsCount=_json.get('itemsCount', None),
                   annotated=_json.get('annotated', None),
                   projects=_json.get('projects', None),
                   creator=_json.get('creator', None),
                   items_url=_json.get('items', None),
                   export=_json.get('export', None),
                   driver=_json.get('driver', None),
                   name=_json.get('name', None),
                   url=_json.get('url', None),
                   id=_json.get('id', None),
                   datasets=datasets,
                   client_api=client_api,
                   project=project)
        inst.is_fetched = is_fetched
        return inst

    def to_json(self):
        """
        Returns platform _json format of object

        :return: platform json format of object
        """
        _json = attr.asdict(self, filter=attr.filters.exclude(attr.fields(Dataset)._client_api,
                                                              attr.fields(Dataset)._project,
                                                              attr.fields(Dataset)._datasets,
                                                              attr.fields(Dataset)._repositories,
                                                              attr.fields(Dataset)._ontology_ids,
                                                              attr.fields(Dataset)._labels,
                                                              attr.fields(Dataset)._directory_tree,
                                                              attr.fields(Dataset)._instance_map,
                                                              attr.fields(Dataset).access_level,
                                                              attr.fields(Dataset).readable_type,
                                                              attr.fields(Dataset).created_at,
                                                              attr.fields(Dataset).items_url))
        _json.update({'items': self.items_url})
        _json['readableType'] = self.readable_type
        _json['createdAt'] = self.created_at
        _json['accessLevel'] = self.access_level
        return _json

    @property
    def labels(self):
        if self._labels is None:
            self._labels = self.recipes.list()[0].ontologies.list()[0].labels
        return self._labels

    @property
    def labels_flat_dict(self):
        flatten_dict = dict()

        def add_to_dict(tag: str, father: entities.Label):
            flatten_dict[tag] = father
            for child in father.children:
                add_to_dict('{}.{}'.format(tag, child.tag), child)

        for label in self.labels:
            add_to_dict(label.tag, label)
        return flatten_dict

    @property
    def instance_map(self):
        if self._instance_map is None:
            labels = [label for label in self.labels_flat_dict]
            labels.sort()
            # each label gets index as instance id
            self._instance_map = {label: (i_label + 1) for i_label, label in enumerate(labels)}
        return self._instance_map

    @instance_map.setter
    def instance_map(self, value: dict):
        """
        instance mapping for creating instance mask
        :param value: dictionary {label: map_id}
        """
        if not isinstance(value, dict):
            raise ValueError('input must be a dictionary of {lable_name: instance_id}')
        self._instance_map = value

    @property
    def ontology_ids(self):
        if self._ontology_ids is None:
            self._ontology_ids = list()
            if self.metadata is not None and 'system' in self.metadata and 'recipes' in self.metadata['system']:
                recipe_ids = self.get_recipe_ids()
                for rec_id in recipe_ids:
                    recipe = self.recipes.get(recipe_id=rec_id)
                    self._ontology_ids += recipe.ontologyIds
        return self._ontology_ids

    @_repositories.default
    def set_repositories(self):
        reps = namedtuple('repositories',
                          field_names=['items', 'recipes', 'datasets', 'assignments', 'tasks', 'annotations'])
        if self._project is None:
            datasets = repositories.Datasets(client_api=self._client_api, project=self._project)
        else:
            datasets = self._project.datasets

        r = reps(items=repositories.Items(client_api=self._client_api, dataset=self, datasets=datasets),
                 recipes=repositories.Recipes(client_api=self._client_api, dataset=self),
                 assignments=repositories.Assignments(project=self._project, client_api=self._client_api, dataset=self),
                 tasks=repositories.Tasks(client_api=self._client_api, project=self._project, dataset=self),
                 annotations=repositories.Annotations(client_api=self._client_api, dataset=self),
                 datasets=datasets)
        return r

    @property
    def items(self):
        assert isinstance(self._repositories.items, repositories.Items)
        return self._repositories.items

    @property
    def recipes(self):
        assert isinstance(self._repositories.recipes, repositories.Recipes)
        return self._repositories.recipes

    @property
    def datasets(self):
        assert isinstance(self._repositories.datasets, repositories.Datasets)
        return self._repositories.datasets

    @property
    def assignments(self):
        assert isinstance(self._repositories.assignments, repositories.Assignments)
        return self._repositories.assignments

    @property
    def tasks(self):
        assert isinstance(self._repositories.tasks, repositories.Tasks)
        return self._repositories.tasks

    @property
    def annotations(self):
        assert isinstance(self._repositories.annotations, repositories.Annotations)
        return self._repositories.annotations

    @property
    def project(self):
        if self._project is None:
            # get from cache
            project = self._client_api.state_io.get('project')
            if project is not None:
                # build entity from json
                p = entities.Project.from_json(_json=project, client_api=self._client_api)
                # check if dataset belongs to project
                if p.id in self.projects:
                    self._project = p
        if self._project is None:
            self._project = repositories.Projects(client_api=self._client_api).get(project_id=self.projects[0],
                                                                                   fetch=None)
        assert isinstance(self._project, entities.Project)
        return self._project

    @project.setter
    def project(self, project):
        if not isinstance(project, entities.Project):
            raise ValueError('Must input a valid Project entity')
        self._project = project

    @property
    def directory_tree(self):
        if self._directory_tree is None:
            self._directory_tree = self.project.datasets.directory_tree(dataset_id=self.id)
        assert isinstance(self._directory_tree, entities.DirectoryTree)
        return self._directory_tree

    def __copy__(self):
        return Dataset.from_json(_json=self.to_json(),
                                 project=self._project,
                                 client_api=self._client_api,
                                 is_fetched=self.is_fetched,
                                 datasets=self.datasets)

    def __get_local_path__(self):
        if self._project is not None:
            local_path = os.path.join(services.service_defaults.DATALOOP_PATH,
                                      'projects',
                                      self.project.name,
                                      'datasets',
                                      self.name)
        else:
            local_path = os.path.join(services.service_defaults.DATALOOP_PATH,
                                      'datasets',
                                      '%s_%s' % (self.name, self.id))
        return local_path

    @staticmethod
    def serialize_labels(labels_dict):
        """
        Convert hex color format to rgb

        :param labels_dict: dict of labels
        :return: dict of converted labels
        """
        dataset_labels_dict = dict()
        for label, color in labels_dict.items():
            dataset_labels_dict[label] = '#%02x%02x%02x' % color
        return dataset_labels_dict

    def get_recipe_ids(self):
        """
        Get dataset recipe Ids

        :return: list of recipe ids
        """
        return self.metadata['system']['recipes']

    def delete(self, sure=False, really=False):
        """
        Delete a dataset forever!

        :param sure: are you sure you want to delete?
        :param really: really really?
        :return:
        """
        return self.datasets.delete(dataset_id=self.id,
                                    sure=sure,
                                    really=really)

    def update(self, system_metadata=False):
        """
        Update dataset field

        :param system_metadata: bool - True, if you want to change metadata system
        :return:
        """
        return self.datasets.update(dataset=self,
                                    system_metadata=system_metadata)

    def clone(self, clone_name, filters=None, with_items_annotations=True, with_metadata=True,
              with_task_annotations_status=True):
        """
        Clone dataset

        :param clone_name: new dataset name
        :param filters: Filters entity or a query dict
        :param with_items_annotations:
        :param with_metadata:
        :param with_task_annotations_status:
        :return:
        """
        return self.datasets.clone(dataset_id=self.id,
                                   filters=filters,
                                   clone_name=clone_name,
                                   with_metadata=with_metadata,
                                   with_items_annotations=with_items_annotations,
                                   with_task_annotations_status=with_task_annotations_status)

    def download_annotations(self,
                             local_path=None,
                             filters=None,
                             annotation_options=None,
                             overwrite=False,
                             thickness=1,
                             with_text=False,
                             remote_path=None,
                             num_workers=32):

        return self.datasets.download_annotations(dataset=self,
                                                  local_path=local_path,
                                                  overwrite=overwrite,
                                                  filters=filters,
                                                  annotation_options=annotation_options,
                                                  thickness=thickness,
                                                  with_text=with_text,
                                                  remote_path=remote_path,
                                                  num_workers=num_workers)

    def checkout(self):
        """
        Checkout the dataset

        :return:
        """
        self.datasets.checkout(dataset=self)

    def open_in_web(self):
        """
        Open the dataset in web platform

        :return:
        """
        self.datasets.open_in_web(dataset=self)

    def add_label(self, label_name, color=None, children=None, attributes=None, display_label=None, label=None,
                  recipe_id=None, ontology_id=None):
        """
        Add single label to dataset

        :param label_name:
        :param color:
        :param children:
        :param attributes:
        :param display_label:
        :param label:
        :param recipe_id: optional
        :param ontology_id: optional
        :return: label entity
        """
        # get recipe
        if recipe_id is None:
            recipe_id = self.get_recipe_ids()[0]
        recipe = self.recipes.get(recipe_id=recipe_id)

        # get ontology
        if ontology_id is None:
            ontology_id = recipe.ontologyIds[0]
        ontology = recipe.ontologies.get(ontology_id=ontology_id)

        # add label
        added_label = ontology.add_label(label_name=label_name,
                                         color=color,
                                         children=children,
                                         attributes=attributes,
                                         display_label=display_label,
                                         label=label)

        # update and return
        ontology = ontology.update(system_metadata=True)
        self._labels = ontology.labels
        return added_label

    def add_labels(self, label_list, ontology_id=None, recipe_id=None):
        """
        Add labels to dataset

        :param label_list:
        :param recipe_id: optional
        :param ontology_id: optional
        :return: label entities
        """
        # get recipe
        if recipe_id is None:
            recipe_id = self.get_recipe_ids()[0]
        recipe = self.recipes.get(recipe_id=recipe_id)

        # get ontology
        if ontology_id is None:
            ontology_id = recipe.ontologyIds[0]
        ontology = recipe.ontologies.get(ontology_id=ontology_id)

        # add labels to ontology
        added_labels = ontology.add_labels(label_list=label_list)

        # update and return
        ontology = ontology.update(system_metadata=True)
        self._labels = ontology.labels
        return added_labels

    def download(
            self,
            filters=None,
            local_path=None,
            file_types=None,
            annotation_options=None,
            overwrite=False,
            to_items_folder=True,
            thickness=1,
            with_text=False,
            without_relative_path=None
    ):
        """
        Download dataset by filters.
        Filtering the dataset for items and save them local
        Optional - also download annotation, mask, instance and image mask of the item

        :param local_path: local folder or filename to save to.
        :param filters: Filters entity or a dictionary containing filters parameters
        :param to_items_folder: Create 'items' folder and download items to it
        :param overwrite: optional - default = False
        :param file_types: a list of file type to download. e.g ['video/webm', 'video/mp4', 'image/jpeg', 'image/png']
        :param annotation_options: download annotations options: dl.ViewAnnotationOptions.list()
        :param with_text: optional - add text to annotations, default = False
        :param thickness: optional - line thickness, if -1 annotation will be filled, default =1
        :param without_relative_path: string - remote path - download items without the relative path from platform
        :return: Output (list)
        """
        return self.items.download(filters=filters,
                                   local_path=local_path,
                                   file_types=file_types,
                                   annotation_options=annotation_options,
                                   overwrite=overwrite,
                                   to_items_folder=to_items_folder,
                                   thickness=thickness,
                                   with_text=with_text,
                                   without_relative_path=without_relative_path)

    def delete_labels(self, label_names):
        """
        Delete labels from dataset's ontologies

        :param label_names: label object/ label name / list of label objects / list of label names
        :return:
        """
        for recipe in self.recipes.list():
            for ontology in recipe.ontologies.list():
                ontology.delete_labels(label_names=label_names)
        self._labels = None



import datetime
import os
from logging import handlers

from .service_defaults import DATALOOP_PATH

class DataloopLogger(handlers.BaseRotatingHandler):
    """
        Based on logging.handlers.RotatingFileHandler
        Create a new log file after reached maxBytes
        Delete logs older than a threshold default is week)
    """

    def __init__(self, filename, mode='a', maxBytes=0, encoding=None, delay=False):
        if maxBytes > 0:
            mode = 'a'
        super().__init__(filename=filename, mode=mode, encoding=encoding, delay=delay)
        self.maxBytes = maxBytes
        DataloopLogger.clean_logs()

    @staticmethod
    def clean_logs():
        log_path = DataloopLogger.get_log_path()
        now = datetime.datetime.utcnow().timestamp()
        files = [os.path.join(log_path, f) for f in os.listdir(log_path)]
        files.sort(key=lambda x: -os.path.getmtime(x))  # newer first
        total_logs_size = 0
        for filepath in files:
            # print(filepath)
            file_time = os.path.getmtime(filepath)
            # delete if older than a week
            if (now - file_time) > (60 * 60 * 24 * 7):  # sec * min * hour * days
                os.remove(filepath)
                continue
            file_size = os.path.getsize(filepath)
            if (total_logs_size + file_size) > 200 * 1e6:  # 100MB
                os.remove(filepath)
                continue
            total_logs_size += file_size

    @staticmethod
    def get_log_path():
        log_path = os.path.join(DATALOOP_PATH, 'logs')
        if not os.path.isdir(log_path):
            os.makedirs(log_path)
        return log_path

    @staticmethod
    def get_log_filepath():
        log_path = DataloopLogger.get_log_path()
        log_filepath = os.path.join(log_path, '{}.log'.format(datetime.datetime.utcnow().strftime('%Y-%m-%d_%H-%M-%S')))
        return log_filepath

    def doRollover(self):
        """
        Do a rollover, as described in __init__().
        """
        if self.stream:
            self.stream.close()
            self.stream = None
        # clean older logs (week old)
        DataloopLogger.clean_logs()
        # create new log
        self.baseFilename = DataloopLogger.get_log_filepath()
        if not self.delay:
            self.stream = self._open()

    def shouldRollover(self, record):
        """
        Determine if rollover should occur.

        Basically, see if the supplied record would cause the file to exceed
        the size limit we have.
        """
        if self.stream is None:  # delay was set...
            self.stream = self._open()
        if self.maxBytes > 0:  # are we rolling over?
            msg = "%s\n" % self.format(record)
            self.stream.seek(0, 2)  # due to non-posix-compliant Windows feature
            if self.stream.tell() + len(msg) >= self.maxBytes:
                return 1
        return 0



import logging

from .. import exceptions, miscellaneous, entities, repositories, services

logger = logging.getLogger(name=__name__)


class Assignments:
    """
    Assignments repository
    """

    def __init__(self,
                 client_api: services.ApiClient,
                 project: entities.Project = None,
                 task: entities.Task = None,
                 dataset: entities.Dataset = None,
                 project_id=None):
        self._client_api = client_api
        self._project = project
        self._dataset = dataset
        self._task = task

        self._project_id = project_id
        if self._project_id is None and self._project is not None:
            self._project_id = self._project.id

    ############
    # entities #
    ############
    @property
    def task(self) -> entities.Task:
        if self._task is None:
            raise exceptions.PlatformException(
                error='2001',
                message='Missing "task". need to set an Task entity or use task.assignments repository')
        assert isinstance(self._task, entities.Task)
        return self._task

    @task.setter
    def task(self, task: entities.Task):
        if not isinstance(task, entities.Task):
            raise ValueError('Must input a valid Task entity')
        self._task = task

    @property
    def project_id(self):
        if self._project_id is not None:
            return self._project_id
        elif self._project is not None:
            return self._project.id
        else:
            return None

    @property
    def project(self) -> entities.Project:
        if self._project is None:
            raise exceptions.PlatformException(
                error='2001',
                message='Missing "project". need to set a Project entity or use project.assignments repository')
        assert isinstance(self._project, entities.Project)
        return self._project

    @project.setter
    def project(self, project: entities.Project):
        if not isinstance(project, entities.Project):
            raise ValueError('Must input a valid Project entity')
        self._project = project

    ###########
    # methods #
    ###########
    def list(self, project_ids=None, status=None, assignment_name=None, assignee_id=None, pages_size=None,
             page_offset=None, task_id=None) -> miscellaneous.List[entities.Assignment]:
        """
        Get Assignments list

        :param task_id:
        :param page_offset:
        :param pages_size:
        :param assignee_id:
        :param assignment_name:
        :param status:
        :param project_ids: list of project ids
        :return: List of Assignment objects
        """

        # url
        url = '/assignments'

        query = list()
        if project_ids is not None:
            if not isinstance(project_ids, list):
                project_ids = [project_ids]
        elif self._project_id is not None:
            project_ids = [self._project_id]
        elif self._project is not None:
            project_ids = [self._project.id]
        else:
            raise ('400', 'Must provide project')

        project_ids = ','.join(project_ids)
        query.append('projects={}'.format(project_ids))

        if status is not None:
            query.append('status={}'.format(status))
        if assignment_name is not None:
            query.append('name={}'.format(assignment_name))
        if assignee_id is not None:
            query.append('annotator={}'.format(assignee_id))
        if pages_size is not None:
            query.append('pageSize={}'.format(pages_size))
        if page_offset is not None:
            query.append('pageOffset={}'.format(page_offset))

        if task_id is None and self._task is not None:
            task_id = self._task.id
        if task_id is not None:
            query.append('taskId={}'.format(task_id))

        if len(query) > 0:
            query_string = '&'.join(query)
            url = '{}?{}'.format(url, query_string)

        success, response = self._client_api.gen_request(req_type='get',
                                                         path=url)
        if success:
            assignments = miscellaneous.List(
                [entities.Assignment.from_json(client_api=self._client_api,
                                               _json=_json, project=self._project, dataset=self._dataset,
                                               task=self._task)
                 for _json in response.json()['items']])
        else:
            logger.error('Platform error getting assignments')
            raise exceptions.PlatformException(response)
        return assignments

    def get(self, assignment_name=None, assignment_id=None):
        """
        Get a Project object
        :param assignment_name: optional - search by name
        :param assignment_id: optional - search by id
        :return: Project object

        """

        if assignment_id is not None:
            url = '/assignments/{}'.format(assignment_id)
            success, response = self._client_api.gen_request(req_type='get',
                                                             path=url)
            if not success:
                raise exceptions.PlatformException('404', 'Assignment not found')
            else:
                assignment = entities.Assignment.from_json(_json=response.json(),
                                                           client_api=self._client_api, project=self._project,
                                                           dataset=self._dataset, task=self._task)
        elif assignment_name is not None:
            assignments = [assignment for assignment in self.list() if assignment.name == assignment_name]
            if len(assignments) == 0:
                raise exceptions.PlatformException('404', 'Assignment not found')
            elif len(assignments) > 1:
                raise exceptions.PlatformException('404',
                                                   'More than one assignment exist with the same name: {}'.format(
                                                       assignment_name))
            else:
                assignment = assignments[0]
        else:
            raise exceptions.PlatformException('400', 'Must provide either assignment name or assignment id')

        assert isinstance(assignment, entities.Assignment)
        return assignment

    def delete(self, assignment: entities.Assignment = None, assignment_name=None, assignment_id=None):
        """
        Delete an assignment
        :param assignment_id:
        :param assignment_name:
        :param assignment:

        :return: True
        """
        if assignment_id is None:
            if assignment is None:
                if assignment_name is None:
                    raise exceptions.PlatformException('400',
                                                       'Must provide either assignment, '
                                                       'assignment name or assignment id')
                else:
                    assignment = self.get(assignment_name=assignment_name)
                    assignment_id = assignment.id
        url = '/assignments/{}'.format(assignment_id)
        success, response = self._client_api.gen_request(req_type='delete',
                                                         path=url)

        if not success:
            raise exceptions.PlatformException(response)
        return True

    def reassign(self, assignee_id, assignment=None, assignment_id=None, task=None, task_id=None):
        """
        Reassign an assignment
        :return: Assignment object
        """
        if assignment_id is None and assignment is None:
            raise exceptions.PlatformException('400', 'Must provide either assignment or assignment_id')
        elif assignment_id is None:
            assignment_id = assignment.id

        if task_id is None and task is None:
            raise exceptions.PlatformException('400', 'Must provide either task or task_id')
        elif task_id is None:
            task_id = task.id

        url = '/assignments/{}/reassign'.format(assignment_id)

        payload = {
            'taskId': task_id,
            'annotator': assignee_id
        }

        success, response = self._client_api.gen_request(req_type='post',
                                                         path=url,
                                                         json_req=payload)
        if success:
            if task is None:
                task = self._task
            return entities.Assignment.from_json(_json=response.json(),
                                                 client_api=self._client_api, project=self._project,
                                                 dataset=self._dataset, task=task)
        else:
            raise exceptions.PlatformException(response)

    def redistribute(self, workload, assignment=None, assignment_id=None, task=None, task_id=None):
        """
        Redistribute an assignment
        :return: Assignment object
        """
        if assignment_id is None and assignment is None:
            raise exceptions.PlatformException('400', 'Must provide either assignment or assignment_id')
        elif assignment_id is None:
            assignment_id = assignment.id

        if task_id is None and task is None:
            raise exceptions.PlatformException('400', 'Must provide either task or task_id')
        elif task_id is None:
            task_id = task.id

        url = '/assignments/{}/redistribute'.format(assignment_id)

        payload = {
            'taskId': task_id,
            'workload': workload.to_json()
        }

        success, response = self._client_api.gen_request(req_type='post',
                                                         path=url,
                                                         json_req=payload)
        if success:
            if task is None:
                task = self._task
            redistributed_assignments = list()
            for redistributed_assignment in response.json():
                redistributed_assignments.append(entities.Assignment.from_json(_json=redistributed_assignment,
                                                                               client_api=self._client_api,
                                                                               project=self._project,
                                                                               dataset=self._dataset, task=task))
            return miscellaneous.List(redistributed_assignments)
        else:
            raise exceptions.PlatformException(response)

    def update(self, assignment: entities.Assignment = None, system_metadata=False) -> entities.Assignment:
        """
        Update an assignment
        :return: Assignment object
        """
        url = '/assignments/{}'.format(assignment.id)

        if system_metadata:
            url += '?system=true'

        success, response = self._client_api.gen_request(req_type='patch',
                                                         path=url,
                                                         json_req=assignment.to_json())
        if success:
            return entities.Assignment.from_json(_json=response.json(),
                                                 client_api=self._client_api, project=self._project,
                                                 dataset=self._dataset, task=self._task)
        else:
            raise exceptions.PlatformException(response)

    def create(self, assignee_id, task=None,
               assignment_name=None,
               dataset=None, project_id=None,
               filters=None, items=None,
               status=None, metadata=None) -> entities.Assignment:
        """
        Create a new assignment
        :param task:
        :param dataset:
        :param items:
        :param filters:
        :param metadata:
        :param project_id:
        :param status:
        :param assignee_id:
        :param assignment_name:
        :return: Assignment object
        """
        if assignment_name:
            logger.warning('DEPRECATED - Param {} will be deprecated after v1.17.0'.format(assignment_name))
        if dataset:
            logger.warning('DEPRECATED - Param {} will be deprecated after v1.17.0'.format(dataset))
        if project_id:
            logger.warning('DEPRECATED - Param {} will be deprecated after v1.17.0'.format(project_id))
        if status:
            logger.warning('DEPRECATED - Param {} will be deprecated after v1.17.0'.format(status))
        if metadata:
            logger.warning('DEPRECATED - Param {} will be deprecated after v1.17.0'.format(metadata))

        return self._create_in_task(assignee_id=assignee_id, task=task, filters=filters, items=items)

    def _create_in_task(self, assignee_id, task, filters=None, items=None) -> entities.Assignment:

        if task is None:
            if self._task is None:
                raise exceptions.PlatformException('400', 'Must provide task')
            task = self._task

        assignments_before = [ass.id for ass in task.assignments.list()]

        if filters is None and items is None:
            raise exceptions.PlatformException('400', 'Must provide either filters or items list')

        workload = entities.Workload.generate(assignee_ids=[assignee_id])
        task = task.add_items(filters=filters, items=items, workload=workload, limit=0)
        assignments = [ass for ass in task.assignments.list() if ass.id not in assignments_before]

        if len(assignments) < 1:
            raise exceptions.PlatformException('Error creating an assignment, '
                                               'Please use task.add_items() to perform this action')

        return assignments[0]

    def __item_operations(self, dataset: entities.Dataset,
                          op, assignment_id=None, assignment_name=None, filters=None, items=None):
        if assignment_id is None and assignment_name is None:
            raise exceptions.PlatformException('400', 'Must provide either assignment name or assignment id')
        elif assignment_id is None:
            assignment_id = self.get(assignment_name=assignment_name).id

        try:
            if filters is None and items is None:
                raise exceptions.PlatformException('400', 'Must provide either filters or items list')

            if filters is None:
                if not isinstance(items, list):
                    items = [items]
                filters = entities.Filters(field='id', values=[item.id for item in items], operator='in')

            filters._ref_assignment = True
            filters._ref_assignment_id = assignment_id
            filters._ref_op = op

            return dataset.items.update(filters=filters)
        finally:
            if filters is not None:
                filters._nullify_refs()

    def assign_items(self, dataset: entities.Dataset,
                     assignment_id=None, assignment_name=None, filters=None, items=None):
        """

        :param assignment_name:
        :param filters:
        :param assignment_id:
        :param dataset:
        :param items:
        :return:
        """
        return self.__item_operations(dataset=dataset, assignment_id=assignment_id, filters=filters, items=items,
                                      op='create', assignment_name=assignment_name)

    def remove_items(self, dataset: entities.Dataset,
                     assignment_id=None, assignment_name=None, filters=None, items=None):
        """

        :param assignment_name:
        :param assignment_id:
        :param dataset:
        :param filters:
        :param items:
        :return:
        """
        return self.__item_operations(dataset=dataset, assignment_id=assignment_id, filters=filters, items=items,
                                      op='delete', assignment_name=assignment_name)

    def get_items(self, assignment: entities.Assignment = None,
                  assignment_id=None, assignment_name=None, dataset=None, filters=None) -> entities.PagedEntities:
        """

        :param filters:
        :param assignment:
        :param dataset:
        :param assignment_id:
        :param assignment_name:
        :return:
        """
        if assignment is None and assignment_id is None and assignment_name is None:
            raise exceptions.PlatformException('400',
                                               'Please provide either assignment,  assignment_id or assignment_name')

        if assignment_id is None:
            if assignment is None:
                assignment = self.get(assignment_name=assignment_name)
            assignment_id = assignment.id

        if dataset is None and self._dataset is None:
            if assignment is None:
                assignment = self.get(assignment_id=assignment_id, assignment_name=assignment_name)
            if assignment.dataset_id is None:
                raise exceptions.PlatformException('400', 'Please provide a dataset entity')
            dataset = repositories.Datasets(client_api=self._client_api, project=self._project).get(
                dataset_id=assignment.dataset_id)
        elif dataset is None:
            dataset = self._dataset

        if filters is None:
            filters = entities.Filters()
        filters.add(field='metadata.system.refs.id', values=[assignment_id], operator='in')

        return dataset.items.list(filters=filters)
